import asyncio
from itertools import chain
from typing import Any, AsyncIterator, Callable, Dict, Optional, Set, Tuple, cast

import pytest
from decoy import Decoy, matchers

from opentrons_hardware.drivers import binary_usb, can_bus
from opentrons_hardware.firmware_bindings import (
    ArbitrationId,
    ArbitrationIdParts,
    MessageId,
)
from opentrons_hardware.firmware_bindings.constants import (
    FirmwareTarget,
    NodeId,
    PipetteName,
    ToolType,
    USBTarget,
)
from opentrons_hardware.firmware_bindings.messages.message_definitions import (
    GetMotorUsageRequest,
    GetMotorUsageResponse,
)
from opentrons_hardware.firmware_bindings.messages.messages import MessageDefinition
from opentrons_hardware.firmware_bindings.messages.payloads import (
    GetMotorUsageResponsePayload,
)
from opentrons_hardware.firmware_bindings.utils import UInt8Field
from opentrons_hardware.firmware_update import FirmwareUpdate, RunUpdate
from opentrons_hardware.firmware_update.types import FirmwareUpdateStatus, StatusElement
from opentrons_hardware.hardware_control import network, tools, types

from opentrons.hardware_control.backends.errors import SubsystemUpdating
from opentrons.hardware_control.backends.subsystem_manager import SubsystemManager
from opentrons.hardware_control.types import SubSystem, SubSystemState


async def _error_update(
    targets: Set[FirmwareTarget],
) -> AsyncIterator[Tuple[FirmwareTarget, StatusElement]]:
    for target in targets:
        yield target, (FirmwareUpdateStatus.queued, 0)
        await asyncio.sleep(0)
    raise RuntimeError("oh no!")


async def _quick_update(
    targets: Set[FirmwareTarget],
) -> AsyncIterator[Tuple[FirmwareTarget, StatusElement]]:
    for target in targets:
        yield target, (FirmwareUpdateStatus.queued, 0)
        await asyncio.sleep(0)
    for value in range(0, 100, 10):
        for target in targets:
            yield target, (FirmwareUpdateStatus.updating, value / 10.0)
            await asyncio.sleep(0)


async def _eternal_update(
    targets: Set[FirmwareTarget],
) -> AsyncIterator[Tuple[FirmwareTarget, StatusElement]]:
    while True:
        for target in targets:
            yield target, (FirmwareUpdateStatus.updating, 0)
            await asyncio.sleep(0)


async def _instant_update(
    targets: Set[FirmwareTarget],
) -> AsyncIterator[Tuple[FirmwareTarget, StatusElement]]:
    for target in targets:
        yield target, (FirmwareUpdateStatus.done, 1.0)
        await asyncio.sleep(0)


def default_subidentifier_for(target: FirmwareTarget) -> int:
    if target in (NodeId.pipette_left, NodeId.pipette_left_bootloader):
        return PipetteName.p1000_single.value
    if target in (NodeId.pipette_right, NodeId.pipette_right_bootloader):
        return PipetteName.p50_multi.value
    return 0


def default_sha() -> str:
    return "abcdef1"


def default_fw_version() -> int:
    return 1


def default_revision() -> types.PCBARevision:
    return types.PCBARevision("A1.0")


def device_info_for(
    target: FirmwareTarget,
    subidentifier: Optional[int] = None,
    version: Optional[int] = None,
) -> network.DeviceInfoCache:
    checked_subidentifier = (
        subidentifier
        if subidentifier is not None
        else default_subidentifier_for(target)
    )
    return network.DeviceInfoCache(
        target=target.application_for(),
        version=version if version is not None else default_fw_version(),
        shortsha=default_sha(),
        flags=None,
        revision=default_revision(),
        subidentifier=checked_subidentifier,
        ok=((not NodeId.is_bootloader(target)) if isinstance(target, NodeId) else True),
    )


def default_network_info_for(
    targets: Set[FirmwareTarget],
) -> Dict[FirmwareTarget, network.DeviceInfoCache]:
    return {target.application_for(): device_info_for(target) for target in targets}


@pytest.fixture
async def can_messenger(decoy: Decoy) -> can_bus.CanMessenger:
    """Build a decoyed can messenger."""
    mock_messenger = decoy.mock(cls=can_bus.CanMessenger)
    listeners: list[Callable[[MessageDefinition, ArbitrationId], None]] = []

    def _add(
        listener: Callable[[MessageDefinition, ArbitrationId], None], _: Any
    ) -> None:
        listeners.append(listener)

    decoy.when(
        mock_messenger.add_listener(  # type: ignore[func-returns-value]
            cast(Any, matchers.Anything()), cast(Any, matchers.Anything())
        )
    ).then_do(_add)

    async def _send_response(node_id: NodeId, message: MessageDefinition) -> None:
        for listener in listeners:
            listener(
                GetMotorUsageResponse(
                    payload=GetMotorUsageResponsePayload(
                        usage_elements=[], num_elements=UInt8Field(0)
                    )
                ),
                ArbitrationId(
                    parts=ArbitrationIdParts(
                        message_id=MessageId.get_motor_usage_response,
                        node_id=NodeId.host,
                        originating_node_id=NodeId.pipette_left,
                    )
                ),
            )
            listener(
                GetMotorUsageResponse(
                    payload=GetMotorUsageResponsePayload(
                        usage_elements=[], num_elements=UInt8Field(0)
                    )
                ),
                ArbitrationId(
                    parts=ArbitrationIdParts(
                        message_id=MessageId.get_motor_usage_response,
                        node_id=NodeId.host,
                        originating_node_id=NodeId.pipette_right,
                    )
                ),
            )
            listener(
                GetMotorUsageResponse(
                    payload=GetMotorUsageResponsePayload(
                        usage_elements=[], num_elements=UInt8Field(0)
                    )
                ),
                ArbitrationId(
                    parts=ArbitrationIdParts(
                        message_id=MessageId.get_motor_usage_response,
                        node_id=NodeId.host,
                        originating_node_id=NodeId.gripper,
                    )
                ),
            )

    decoy.when(
        await mock_messenger.send(  # type: ignore[func-returns-value]
            node_id=NodeId.broadcast, message=GetMotorUsageRequest()
        )
    ).then_do(_send_response)
    return mock_messenger


@pytest.fixture
def usb_messenger(decoy: Decoy) -> binary_usb.BinaryMessenger:
    """Build a decoyed USB messenger."""
    return decoy.mock(cls=binary_usb.BinaryMessenger)


@pytest.fixture
def tool_detector(decoy: Decoy) -> tools.detector.ToolDetector:
    return decoy.mock(cls=tools.detector.ToolDetector)


@pytest.fixture
def network_info(decoy: Decoy) -> network.NetworkInfo:
    return decoy.mock(cls=network.NetworkInfo)


@pytest.fixture
def update_bag(decoy: Decoy) -> FirmwareUpdate:
    return decoy.mock(cls=FirmwareUpdate)


@pytest.fixture
def detection_queue(
    tool_detector: tools.detector.ToolDetector, decoy: Decoy
) -> "asyncio.Queue[tools.types.ToolDetectionResult]":
    queue: "asyncio.Queue[tools.types.ToolDetectionResult]" = asyncio.Queue()

    async def _read() -> AsyncIterator[tools.types.ToolDetectionResult]:
        while True:
            yield await queue.get()

    decoy.when(tool_detector.detect()).then_return(_read())
    return queue


def prep_mock_update(
    update_bag: FirmwareUpdate, decoy: Decoy, update_details: Dict[FirmwareTarget, str]
) -> RunUpdate:
    updater = decoy.mock(cls=RunUpdate)
    update_class = update_bag.update_runner
    decoy.when(
        update_class(
            can_messenger=matchers.Anything(),
            usb_messenger=matchers.Anything(),
            update_details=update_details,
            retry_count=matchers.Anything(),
            timeout_seconds=matchers.Anything(),
            erase=matchers.Anything(),
        )
    ).then_return(updater)
    return updater


class ToolDetectionController:
    def __init__(
        self,
        tool_detector: tools.detector.ToolDetector,
        detection_queue: "asyncio.Queue[tools.types.ToolDetectionResult]",
        decoy: Decoy,
    ) -> None:
        self._tool_detector = tool_detector
        self._decoy = decoy
        self._detection_queue = detection_queue

    def _detection_for(
        self, targets: Set[FirmwareTarget]
    ) -> tools.types.ToolDetectionResult:
        left_tool = ToolType.nothing_attached
        right_tool = ToolType.nothing_attached
        gripper_tool = ToolType.nothing_attached
        if NodeId.pipette_left in targets or NodeId.pipette_left_bootloader in targets:
            left_tool = ToolType.pipette
        if NodeId.pipette_right in targets or NodeId.pipette_left_bootloader in targets:
            right_tool = ToolType.pipette
        if NodeId.gripper in targets or NodeId.gripper_bootloader in targets:
            gripper_tool = ToolType.gripper
        return tools.types.ToolDetectionResult(
            left=left_tool, right=right_tool, gripper=gripper_tool
        )

    async def add_detection_on_next_check(
        self, targets: Set[FirmwareTarget]
    ) -> tools.types.ToolDetectionResult:
        results = self._detection_for(targets)

        async def _adder() -> None:
            await self._detection_queue.put(results)

        self._decoy.when(await self._tool_detector.check_once()).then_do(_adder)  # type: ignore[func-returns-value]
        return results

    async def add_detection_immediate(
        self,
        targets: Set[FirmwareTarget],
    ) -> tools.types.ToolDetectionResult:
        result = self._detection_for(targets)
        await self._detection_queue.put(result)
        return result

    def _pipette_info_from_network(
        self, device: Optional[network.DeviceInfoCache], default_name: PipetteName
    ) -> Optional[tools.types.PipetteInformation]:
        if not device:
            return None
        if not device.ok:
            return None
        pipette_name = (
            PipetteName(device.subidentifier)
            if device.subidentifier != 0
            else default_name
        )
        return tools.types.PipetteInformation(
            name=pipette_name,
            name_int=pipette_name.value,
            model=pipette_name.name,
            serial=f"dummyserial{pipette_name.name}",
        )

    def _auto_tool_summary(
        self, devices: Dict[FirmwareTarget, network.DeviceInfoCache]
    ) -> tools.types.ToolSummary:
        return tools.types.ToolSummary(
            left=self._pipette_info_from_network(
                devices.get(NodeId.pipette_left), PipetteName.p1000_single
            ),
            right=self._pipette_info_from_network(
                devices.get(NodeId.pipette_right), PipetteName.p50_multi
            ),
            gripper=(
                tools.types.GripperInformation(model="1", serial="12131231")
                if (NodeId.gripper in devices and devices[NodeId.gripper].ok)
                else None
            ),
        )

    async def add_resolution(
        self,
        on: tools.types.ToolDetectionResult,
        devices: Dict[FirmwareTarget, network.DeviceInfoCache],
        specific: Optional[tools.types.ToolSummary] = None,
    ) -> tools.types.ToolSummary:
        summary = specific or self._auto_tool_summary(devices)

        arg = tools.types.ToolDetectionResult(
            left=(
                on.left
                if (NodeId.pipette_left in devices and devices[NodeId.pipette_left].ok)
                else ToolType.nothing_attached
            ),
            right=(
                on.right
                if (
                    NodeId.pipette_right in devices and devices[NodeId.pipette_right].ok
                )
                else ToolType.nothing_attached
            ),
            gripper=(
                on.gripper
                if (NodeId.gripper in devices and devices[NodeId.gripper].ok)
                else ToolType.nothing_attached
            ),
        )

        self._decoy.when(await self._tool_detector.resolve(arg, 10.0)).then_return(
            summary
        )
        return summary


@pytest.fixture
def tool_detection_controller(
    tool_detector: tools.detector.ToolDetector,
    detection_queue: "asyncio.Queue[tools.types.ToolDetectionResult]",
    decoy: Decoy,
) -> ToolDetectionController:
    return ToolDetectionController(tool_detector, detection_queue, decoy)


@pytest.fixture
async def subject(
    can_messenger: can_bus.CanMessenger,
    usb_messenger: binary_usb.BinaryMessenger,
    tool_detector: tools.detector.ToolDetector,
    network_info: network.NetworkInfo,
    update_bag: FirmwareUpdate,
) -> AsyncIterator[SubsystemManager]:
    """Build a test subject using decoyed messengers."""
    manager = SubsystemManager(
        can_messenger, usb_messenger, tool_detector, network_info, update_bag
    )
    try:
        yield manager
    finally:
        if manager._tool_detection_task:
            manager._tool_detection_task.cancel()
            with pytest.raises(asyncio.CancelledError):
                await manager._tool_detection_task


@pytest.mark.parametrize(
    "targets,ok",
    [
        (set(), False),  # with no targets, we're not ok
        (
            {NodeId.head, NodeId.gantry_x, NodeId.gantry_y},
            False,
        ),  # with all can targets but no usb we're not ok
        (
            {NodeId.head, NodeId.gantry_x, NodeId.gantry_y, USBTarget.rear_panel},
            True,
        ),  # with all targets we're ok
        (
            {
                NodeId.head,
                NodeId.gantry_x,
                NodeId.gantry_y_bootloader,
                USBTarget.rear_panel,
            },
            False,
        ),  # with a core can node in bootloader we're not ok
        (
            {
                NodeId.head,
                NodeId.gantry_x,
                NodeId.gantry_y,
                USBTarget.rear_panel,
                NodeId.pipette_right,
            },
            True,
        ),  # with all plus a pipette we're ok
        (
            {
                NodeId.head_bootloader,
                NodeId.gantry_x,
                NodeId.gantry_y,
                USBTarget.rear_panel,
                NodeId.pipette_right,
            },
            False,
        ),  # with a non-ok core and an ok pipette we're not ok
        (
            {
                NodeId.head,
                NodeId.gantry_x,
                NodeId.gantry_y,
                USBTarget.rear_panel,
                NodeId.gripper_bootloader,
            },
            False,
        ),  # with an ok core but not ok tool we're not ok
    ],
)
async def test_ok(
    targets: Set[FirmwareTarget],
    ok: bool,
    subject: SubsystemManager,
    network_info: network.NetworkInfo,
    tool_detection_controller: ToolDetectionController,
    update_bag: FirmwareUpdate,
    decoy: Decoy,
) -> None:
    """It should complain if required subsystems are not present."""
    target_applications = {t.application_for() for t in targets}
    decoy.when(network_info.targets).then_return(target_applications)
    network_info_value = default_network_info_for(targets)
    decoy.when(network_info.device_info).then_return(network_info_value)
    tool_struct = await tool_detection_controller.add_detection_on_next_check(targets)
    await tool_detection_controller.add_resolution(tool_struct, network_info_value)

    good_targets = {
        target.application_for()
        for target in targets
        if not (isinstance(target, NodeId) and NodeId.is_bootloader(target))
    }
    bad_targets = target_applications - good_targets
    decoy.when(
        update_bag.update_checker(network_info_value, good_targets, False)
    ).then_return({})
    decoy.when(
        update_bag.update_checker(
            network_info_value,
            bad_targets,
            True,
        )
    ).then_return({target: (1, "/some/path") for target in bad_targets})
    await subject.start()
    assert subject.ok == ok


async def test_device_info(
    subject: SubsystemManager, network_info: network.NetworkInfo, decoy: Decoy
) -> None:
    """It should accurately reflect the device info data from the network."""
    info_value = default_network_info_for(
        {
            NodeId.gantry_x,
            NodeId.gantry_y_bootloader,
            NodeId.head,
            NodeId.pipette_right,
            USBTarget.rear_panel,
        }
    )
    decoy.when(network_info.device_info).then_return(info_value)
    assert subject.device_info == {
        SubSystem.gantry_x: info_value[NodeId.gantry_x],
        SubSystem.gantry_y: info_value[NodeId.gantry_y],
        SubSystem.head: info_value[NodeId.head],
        SubSystem.pipette_right: info_value[NodeId.pipette_right],
        SubSystem.rear_panel: info_value[USBTarget.rear_panel],
    }


async def test_targets(
    subject: SubsystemManager, network_info: network.NetworkInfo, decoy: Decoy
) -> None:
    """It should tell you what targets are connected if you really need to know."""
    targets: Set[FirmwareTarget] = {
        NodeId.gantry_x,
        NodeId.gantry_y_bootloader,
        USBTarget.rear_panel,
        NodeId.gripper,
    }
    decoy.when(network_info.targets).then_return(targets)
    assert subject.targets == targets


@pytest.mark.parametrize(
    "target_info",
    [
        {},
        {
            NodeId.pipette_left: device_info_for(NodeId.pipette_left, version=0),
            NodeId.pipette_right: device_info_for(NodeId.pipette_right),
        },
        {NodeId.gripper: device_info_for(NodeId.gripper)},
        {
            NodeId.gripper: device_info_for(NodeId.gripper, version=0),
            NodeId.pipette_left: device_info_for(
                NodeId.pipette_left, subidentifier=PipetteName.p1000_96.value
            ),
        },
    ],
)
async def test_ok_tools_get_mapped(
    target_info: Dict[FirmwareTarget, network.DeviceInfoCache],
    subject: SubsystemManager,
    network_info: network.NetworkInfo,
    tool_detection_controller: ToolDetectionController,
    update_bag: FirmwareUpdate,
    decoy: Decoy,
) -> None:
    """When tools are detected no matter their update needs, if they're ok we make them present."""
    core_targets: Set[FirmwareTarget] = {
        NodeId.gantry_x,
        NodeId.gantry_y,
        NodeId.head,
        USBTarget.rear_panel,
    }
    all_targets = set(iter(target_info.keys())).union(core_targets)
    decoy.when(network_info.targets).then_return(all_targets)
    core_network_info = default_network_info_for(core_targets)
    all_network_info = dict(chain(core_network_info.items(), target_info.items()))
    decoy.when(network_info.device_info).then_return(all_network_info)

    tool_struct = await tool_detection_controller.add_detection_on_next_check(
        set(iter(target_info.keys()))
    )
    summary = await tool_detection_controller.add_resolution(tool_struct, target_info)

    decoy.when(
        update_bag.update_checker(all_network_info, all_targets, False)
    ).then_return(
        {
            target: (1, "/some/path")
            for target in target_info.keys()
            if target.is_bootloader()
        }
    )
    decoy.when(
        update_bag.update_checker(
            all_network_info,
            set(),
            True,
        )
    ).then_return({})
    await subject.start()
    assert subject.tools == summary


@pytest.mark.parametrize(
    "target_info,subsystem_info,required_updates",
    [
        ({}, {}, {}),
        (
            default_network_info_for({NodeId.head_bootloader, USBTarget.rear_panel}),
            {
                SubSystem.head: SubSystemState(
                    ok=False,
                    current_fw_version=default_fw_version(),
                    next_fw_version=1,
                    fw_update_needed=True,
                    current_fw_sha=default_sha(),
                    pcba_revision=str(default_revision()),
                    update_state=None,
                ),
                SubSystem.rear_panel: SubSystemState(
                    ok=True,
                    current_fw_version=default_fw_version(),
                    next_fw_version=2,
                    fw_update_needed=True,
                    current_fw_sha=default_sha(),
                    pcba_revision=str(default_revision()),
                    update_state=None,
                ),
            },
            {
                NodeId.head: (1, "/some/path"),
                USBTarget.rear_panel: (2, "/some/other/path"),
            },
        ),
        (
            default_network_info_for({NodeId.pipette_left, NodeId.gantry_x}),
            {
                SubSystem.pipette_left: SubSystemState(
                    ok=True,
                    current_fw_version=default_fw_version(),
                    next_fw_version=default_fw_version(),
                    fw_update_needed=False,
                    current_fw_sha=default_sha(),
                    pcba_revision=str(default_revision()),
                    update_state=None,
                ),
                SubSystem.gantry_x: SubSystemState(
                    ok=True,
                    current_fw_version=default_fw_version(),
                    next_fw_version=default_fw_version(),
                    fw_update_needed=False,
                    current_fw_sha=default_sha(),
                    pcba_revision=str(default_revision()),
                    update_state=None,
                ),
            },
            {},
        ),
        (
            default_network_info_for(
                {NodeId.head_bootloader, NodeId.pipette_right_bootloader}
            ),
            {
                SubSystem.head: SubSystemState(
                    ok=False,
                    current_fw_version=default_fw_version(),
                    next_fw_version=default_fw_version(),
                    fw_update_needed=True,
                    current_fw_sha=default_sha(),
                    pcba_revision=str(default_revision()),
                    update_state=None,
                ),
                SubSystem.pipette_right: SubSystemState(
                    ok=False,
                    current_fw_version=default_fw_version(),
                    next_fw_version=default_fw_version(),
                    fw_update_needed=True,
                    current_fw_sha=default_sha(),
                    pcba_revision=str(default_revision()),
                    update_state=None,
                ),
            },
            {NodeId.head: (1, "/some/path"), NodeId.pipette_right: (1, "/some/path")},
        ),
    ],
)
async def test_subsystems(
    target_info: Dict[FirmwareTarget, network.DeviceInfoCache],
    subsystem_info: Dict[SubSystem, SubSystemState],
    required_updates: Dict[FirmwareTarget, Tuple[int, str]],
    subject: SubsystemManager,
    network_info: network.NetworkInfo,
    tool_detection_controller: ToolDetectionController,
    update_bag: FirmwareUpdate,
    decoy: Decoy,
) -> None:
    """It should return subsystems corresponding to the detected targets with their info."""
    targets = set(iter(target_info.keys()))
    target_applications = {t.application_for() for t in target_info.keys()}
    decoy.when(network_info.targets).then_return(target_applications)
    decoy.when(network_info.device_info).then_return(target_info)
    tool_struct = await tool_detection_controller.add_detection_on_next_check(targets)
    await tool_detection_controller.add_resolution(tool_struct, target_info)

    good_targets = {target for target, info in target_info.items() if info.ok}
    bad_targets = target_applications - good_targets
    decoy.when(update_bag.update_checker(target_info, good_targets, False)).then_return(
        {k: v for k, v in required_updates.items() if k in good_targets}
    )
    decoy.when(
        update_bag.update_checker(
            target_info,
            bad_targets,
            True,
        )
    ).then_return({k: v for k, v in required_updates.items() if k in bad_targets})
    await subject.start()
    assert subject.subsystems == subsystem_info


@pytest.mark.parametrize(
    "target_info,required_updates,outcome",
    [
        ({}, {}, False),
        (
            default_network_info_for({NodeId.head_bootloader, USBTarget.rear_panel}),
            {
                NodeId.head: (1, "/some/path"),
                USBTarget.rear_panel: (2, "/some/other/path"),
            },
            True,
        ),
        (
            default_network_info_for({NodeId.pipette_left, NodeId.gantry_x}),
            {},
            False,
        ),
        (
            default_network_info_for(
                {NodeId.head_bootloader, NodeId.pipette_right_bootloader}
            ),
            {NodeId.head: (1, "/some/path"), NodeId.pipette_right: (1, "/some/path")},
            True,
        ),
    ],
)
async def test_update_required(
    target_info: Dict[FirmwareTarget, network.DeviceInfoCache],
    required_updates: Dict[FirmwareTarget, Tuple[int, str]],
    outcome: bool,
    subject: SubsystemManager,
    network_info: network.NetworkInfo,
    tool_detection_controller: ToolDetectionController,
    update_bag: FirmwareUpdate,
    decoy: Decoy,
) -> None:
    """It should have an accurate update_required attribute corresponding to whether updates are needed.."""
    targets = set(iter(target_info.keys()))
    target_applications = {t.application_for() for t in target_info.keys()}
    decoy.when(network_info.targets).then_return(target_applications)
    decoy.when(network_info.device_info).then_return(target_info)
    tool_struct = await tool_detection_controller.add_detection_on_next_check(targets)
    await tool_detection_controller.add_resolution(tool_struct, target_info)

    good_targets = {target for target, info in target_info.items() if info.ok}
    bad_targets = target_applications - good_targets
    decoy.when(update_bag.update_checker(target_info, good_targets, False)).then_return(
        {k: v for k, v in required_updates.items() if k in good_targets}
    )
    decoy.when(
        update_bag.update_checker(
            target_info,
            bad_targets,
            True,
        )
    ).then_return({k: v for k, v in required_updates.items() if k in bad_targets})
    await subject.start()
    assert subject.update_required == outcome


@pytest.mark.parametrize(
    "present,subsystems,required_updates",
    [
        ({}, set(), {}),
        (
            default_network_info_for({NodeId.head, USBTarget.rear_panel}),
            {SubSystem.gantry_x},
            {},
        ),
        (
            default_network_info_for(
                {NodeId.head, NodeId.gantry_x, NodeId.gantry_y, USBTarget.rear_panel}
            ),
            {SubSystem.gripper},
            {},
        ),
        (default_network_info_for({NodeId.head}), {SubSystem.head}, {}),
    ],
)
@pytest.mark.parametrize("force", [False, True])
async def test_update_does_not_happen_for_missing_subsystem(
    present: Dict[FirmwareTarget, network.DeviceInfoCache],
    subsystems: Optional[Set[SubSystem]],
    required_updates: Dict[FirmwareTarget, Tuple[int, str]],
    force: bool,
    subject: SubsystemManager,
    update_bag: FirmwareUpdate,
    network_info: network.NetworkInfo,
    decoy: Decoy,
) -> None:
    """It should instantly complete the update of missing subsystems."""
    targets = set(iter(present.keys()))
    decoy.when(network_info.device_info).then_return(present)
    decoy.when(network_info.targets).then_return(targets)
    decoy.when(
        update_bag.update_checker(
            matchers.Anything(), matchers.Anything(), matchers.Anything()
        )
    ).then_return(required_updates)
    update_generator = subject.update_firmware(subsystems, force)
    with pytest.raises(StopAsyncIteration):
        await update_generator.__anext__()


async def test_exception_on_multiple_updates(
    subject: SubsystemManager,
    update_bag: FirmwareUpdate,
    network_info: network.NetworkInfo,
    decoy: Decoy,
) -> None:
    """If you update a subsystem when an update is already occurring you get an exception."""
    decoy.when(network_info.device_info).then_return(
        default_network_info_for({NodeId.pipette_right})
    )
    decoy.when(network_info.targets).then_return({NodeId.pipette_right})
    decoy.when(
        update_bag.update_checker(
            matchers.Anything(), {NodeId.pipette_right}, matchers.Anything()
        )
    ).then_return({NodeId.pipette_right: (1, "/some/path")})
    decoy.when(update_bag.update_checker(matchers.Anything(), set(), True)).then_return(
        {}
    )

    other_updater = prep_mock_update(
        update_bag, decoy, {NodeId.pipette_right: "/some/path"}
    )
    decoy.when(other_updater.run_updates()).then_return(
        _eternal_update({NodeId.pipette_right})
    )

    running = asyncio.Event()

    async def _fake_update() -> None:
        async for _ in subject.update_firmware({SubSystem.pipette_right}):
            await asyncio.sleep(0)
            running.set()

    other_update = asyncio.create_task(_fake_update())
    try:
        await asyncio.wait_for(running.wait(), 0.1)
    except asyncio.TimeoutError:
        other_update.result()
    while not subject.get_update_progress():
        await asyncio.sleep(0)
    new_updater = prep_mock_update(
        update_bag, decoy, {NodeId.pipette_right: "/some/path"}
    )
    decoy.when(new_updater.run_updates()).then_return(
        _quick_update({NodeId.pipette_right})
    )

    try:
        with pytest.raises(SubsystemUpdating):
            async for _ in subject.update_firmware({SubSystem.pipette_right}):
                await asyncio.sleep(0)
    finally:
        other_update.cancel()


async def test_update_succeeds(
    subject: SubsystemManager,
    update_bag: FirmwareUpdate,
    network_info: network.NetworkInfo,
    decoy: Decoy,
    tool_detection_controller: ToolDetectionController,
) -> None:
    """It should successfully run the update generator."""
    targets: Set[FirmwareTarget] = {
        NodeId.pipette_right,
        NodeId.gantry_x,
        NodeId.gantry_y,
        USBTarget.rear_panel,
        NodeId.head,
    }
    network_info_value = default_network_info_for(targets)
    decoy.when(network_info.device_info).then_return(network_info_value)
    decoy.when(network_info.targets).then_return(targets)

    decoy.when(
        update_bag.update_checker(matchers.Anything(), matchers.Anything(), False)
    ).then_return(
        {
            NodeId.pipette_right: (1, "/some/path"),
            NodeId.gantry_x: (1, "/some/other/path"),
        }
    )
    decoy.when(update_bag.update_checker(matchers.Anything(), set(), True)).then_return(
        {}
    )

    updater = prep_mock_update(
        update_bag,
        decoy,
        {NodeId.pipette_right: "/some/path", NodeId.gantry_x: "/some/other/path"},
    )
    decoy.when(updater.run_updates()).then_return(
        _quick_update({NodeId.pipette_right, NodeId.gantry_x})
    )
    tool_struct = await tool_detection_controller.add_detection_on_next_check(targets)
    await tool_detection_controller.add_resolution(tool_struct, network_info_value)

    await subject.start()

    async for status in subject.update_firmware(
        {SubSystem.pipette_right, SubSystem.gantry_x}
    ):
        assert status.subsystem in {SubSystem.pipette_right, SubSystem.gantry_x}
        assert status.subsystem in subject.get_update_progress()

    assert subject.get_update_progress() == {}


async def test_update_process_fails(
    subject: SubsystemManager,
    update_bag: FirmwareUpdate,
    network_info: network.NetworkInfo,
    decoy: Decoy,
    tool_detection_controller: ToolDetectionController,
) -> None:
    """If the update coroutine fails, the error should be propagated."""
    targets: Set[FirmwareTarget] = {
        NodeId.pipette_right,
        NodeId.gantry_x,
        NodeId.gantry_y,
        USBTarget.rear_panel,
        NodeId.head,
    }
    network_info_value = default_network_info_for(targets)
    decoy.when(network_info.device_info).then_return(network_info_value)
    decoy.when(network_info.targets).then_return(targets)
    decoy.when(
        update_bag.update_checker(matchers.Anything(), matchers.Anything(), False)
    ).then_return({NodeId.pipette_right: (1, "/some/path")})
    decoy.when(update_bag.update_checker(matchers.Anything(), set(), True)).then_return(
        {}
    )

    updater = prep_mock_update(update_bag, decoy, {NodeId.pipette_right: "/some/path"})
    decoy.when(updater.run_updates()).then_return(_error_update({NodeId.pipette_right}))
    tool_struct = await tool_detection_controller.add_detection_on_next_check(targets)
    await tool_detection_controller.add_resolution(tool_struct, network_info_value)

    await subject.start()
    with pytest.raises(RuntimeError):
        async for status in subject.update_firmware(
            {SubSystem.pipette_right, SubSystem.gantry_x}
        ):
            assert status.subsystem in {SubSystem.pipette_right, SubSystem.gantry_x}
            assert status.subsystem in subject.get_update_progress()

    assert subject.get_update_progress() == {}


async def test_update_process_fails_if_subsytem_does_not_appear_afterward(
    subject: SubsystemManager,
    update_bag: FirmwareUpdate,
    network_info: network.NetworkInfo,
    decoy: Decoy,
    tool_detection_controller: ToolDetectionController,
) -> None:
    """After the update process, the updated subsystem must appear ok."""
    targets: Set[FirmwareTarget] = {
        NodeId.pipette_right,
        NodeId.gantry_x,
        NodeId.gantry_y,
        USBTarget.rear_panel,
        NodeId.head,
    }
    network_info_value = default_network_info_for(targets)
    decoy.when(network_info.device_info).then_return(network_info_value)
    decoy.when(network_info.targets).then_return(targets)
    decoy.when(
        update_bag.update_checker(matchers.Anything(), matchers.Anything(), False)
    ).then_return({NodeId.gantry_x: (1, "/some/path")})
    decoy.when(update_bag.update_checker(matchers.Anything(), set(), True)).then_return(
        {}
    )

    updater = prep_mock_update(update_bag, decoy, {NodeId.gantry_x: "/some/path"})
    decoy.when(updater.run_updates()).then_return(_quick_update({NodeId.gantry_x}))
    tool_struct = await tool_detection_controller.add_detection_on_next_check(targets)
    await tool_detection_controller.add_resolution(tool_struct, network_info_value)

    await subject.start()
    subject._check_device_update_timeout = 0.5
    iteration = 0
    with pytest.raises(RuntimeError, match="Device failed"):
        async for status in subject.update_firmware({SubSystem.gantry_x}):
            iteration += 1
            if iteration == 3:
                new_targets = targets - {NodeId.gantry_x}
                decoy.when(network_info.device_info).then_return(
                    default_network_info_for(new_targets)
                )
                decoy.when(network_info.targets).then_return(new_targets)

            assert status.subsystem == SubSystem.gantry_x
            assert status.subsystem in subject.get_update_progress()

    assert subject.get_update_progress() == {}


async def test_update_process_fails_if_subsytem_is_not_ok_afterward(
    subject: SubsystemManager,
    update_bag: FirmwareUpdate,
    network_info: network.NetworkInfo,
    decoy: Decoy,
    tool_detection_controller: ToolDetectionController,
) -> None:
    """After the update process, the updated subsystem must appear ok."""
    targets: Set[FirmwareTarget] = {
        NodeId.pipette_right,
        NodeId.gantry_x,
        NodeId.gantry_y,
        USBTarget.rear_panel,
        NodeId.head,
    }
    network_info_value = default_network_info_for(targets)
    decoy.when(network_info.device_info).then_return(network_info_value)
    decoy.when(network_info.targets).then_return(targets)
    decoy.when(
        update_bag.update_checker(matchers.Anything(), matchers.Anything(), False)
    ).then_return({NodeId.gantry_x: (1, "/some/path")})
    decoy.when(update_bag.update_checker(matchers.Anything(), set(), True)).then_return(
        {}
    )

    updater = prep_mock_update(update_bag, decoy, {NodeId.gantry_x: "/some/path"})
    decoy.when(updater.run_updates()).then_return(_quick_update({NodeId.gantry_x}))
    tool_struct = await tool_detection_controller.add_detection_on_next_check(targets)
    await tool_detection_controller.add_resolution(tool_struct, network_info_value)

    await subject.start()
    subject._check_device_update_timeout = 0.5
    iteration = 0
    with pytest.raises(RuntimeError, match="Device failed"):
        async for status in subject.update_firmware({SubSystem.gantry_x}):
            iteration += 1
            if iteration == 3:
                new_targets = (targets - {NodeId.gantry_x}).union(
                    {NodeId.gantry_x_bootloader}
                )
                decoy.when(network_info.device_info).then_return(
                    default_network_info_for(new_targets)
                )
                decoy.when(network_info.targets).then_return(
                    {x.application_for() for x in new_targets}
                )
                decoy.when(
                    update_bag.update_checker(
                        matchers.Anything(), {NodeId.gantry_x}, True
                    )
                ).then_return({NodeId.gantry_x: (1, "/some/path")})

            assert status.subsystem == SubSystem.gantry_x
            assert status.subsystem in subject.get_update_progress()

    assert subject.get_update_progress() == {}


async def test_updates_update_no_more_than_speciied(
    subject: SubsystemManager,
    update_bag: FirmwareUpdate,
    network_info: network.NetworkInfo,
    decoy: Decoy,
    tool_detection_controller: ToolDetectionController,
) -> None:
    """Even if more subsystems require updates, only arguments should be updated."""
    targets: Set[FirmwareTarget] = {
        NodeId.pipette_right,
        NodeId.gantry_x,
        NodeId.gantry_y,
        USBTarget.rear_panel,
        NodeId.head,
    }
    network_info_value = default_network_info_for(targets)
    decoy.when(network_info.device_info).then_return(network_info_value)
    decoy.when(network_info.targets).then_return(targets)

    decoy.when(
        update_bag.update_checker(matchers.Anything(), matchers.Anything(), False)
    ).then_return(
        {
            NodeId.pipette_right: (1, "/some/path"),
            NodeId.gantry_x: (1, "/some/other/path"),
        }
    )
    decoy.when(update_bag.update_checker(matchers.Anything(), set(), True)).then_return(
        {}
    )

    updater = prep_mock_update(
        update_bag,
        decoy,
        {NodeId.pipette_right: "/some/path"},
    )
    decoy.when(updater.run_updates()).then_return(_quick_update({NodeId.pipette_right}))
    tool_struct = await tool_detection_controller.add_detection_on_next_check(targets)
    await tool_detection_controller.add_resolution(tool_struct, network_info_value)

    await subject.start()

    async for status in subject.update_firmware({SubSystem.pipette_right}):
        assert status.subsystem == SubSystem.pipette_right
        assert status.subsystem in subject.get_update_progress()

    assert subject.get_update_progress() == {}



"""Test update-position-estimator commands."""

from decoy import Decoy

from opentrons.hardware_control import OT3HardwareControlAPI
from opentrons.hardware_control.types import Axis
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.unsafe.unsafe_engage_axes import (
    UnsafeEngageAxesImplementation,
    UnsafeEngageAxesParams,
    UnsafeEngageAxesResult,
)
from opentrons.protocol_engine.execution import GantryMover
from opentrons.protocol_engine.types import MotorAxis


async def test_engage_axes_implementation(
    decoy: Decoy, ot3_hardware_api: OT3HardwareControlAPI, gantry_mover: GantryMover
) -> None:
    """Test EngageAxes command execution."""
    subject = UnsafeEngageAxesImplementation(
        hardware_api=ot3_hardware_api, gantry_mover=gantry_mover
    )

    data = UnsafeEngageAxesParams(
        axes=[
            MotorAxis.LEFT_Z,
            MotorAxis.LEFT_PLUNGER,
            MotorAxis.X,
            MotorAxis.Y,
            MotorAxis.RIGHT_Z,
            MotorAxis.RIGHT_PLUNGER,
        ]
    )
    decoy.when(
        gantry_mover.motor_axes_to_present_hardware_axes(
            [
                MotorAxis.LEFT_Z,
                MotorAxis.LEFT_PLUNGER,
                MotorAxis.X,
                MotorAxis.Y,
                MotorAxis.RIGHT_Z,
                MotorAxis.RIGHT_PLUNGER,
            ]
        )
    ).then_return([Axis.Z_L, Axis.P_L, Axis.X, Axis.Y])

    decoy.when(
        await ot3_hardware_api.update_axis_position_estimations(  # type: ignore[func-returns-value]
            [Axis.Z_L, Axis.P_L, Axis.X, Axis.Y]
        )
    ).then_return(None)

    result = await subject.execute(data)

    assert result == SuccessData(public=UnsafeEngageAxesResult())

    decoy.verify(
        await ot3_hardware_api.engage_axes([Axis.Z_L, Axis.P_L, Axis.X, Axis.Y]),
    )



"""Test configure nozzle layout commands."""

from collections import OrderedDict
from typing import Dict, Union

import pytest
from decoy import Decoy

from opentrons_shared_data.pipette.pipette_definition import (
    AvailableSensorDefinition,
    ValidNozzleMaps,
)
from opentrons_shared_data.pipette.types import (
    LiquidClasses as VolumeModes,
)
from opentrons_shared_data.pipette.types import (
    PipetteNameType,
)

import opentrons.protocol_engine.state.update_types as update_types
from ..pipette_fixtures import (
    NINETY_SIX_COLS,
    NINETY_SIX_MAP,
    NINETY_SIX_ROWS,
    get_default_nozzle_map,
)
from opentrons.hardware_control.nozzle_manager import NozzleMap
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.configure_nozzle_layout import (
    ConfigureNozzleLayoutImplementation,
    ConfigureNozzleLayoutParams,
    ConfigureNozzleLayoutResult,
)
from opentrons.protocol_engine.execution import (
    EquipmentHandler,
    LoadedConfigureNozzleLayoutData,
    TipHandler,
)
from opentrons.protocol_engine.resources.pipette_data_provider import (
    LoadedStaticPipetteData,
)
from opentrons.protocol_engine.state.update_types import (
    PipetteNozzleMapUpdate,
    StateUpdate,
)
from opentrons.protocol_engine.types import (
    AllNozzleLayoutConfiguration,
    ColumnNozzleLayoutConfiguration,
    FlowRates,
    QuadrantNozzleLayoutConfiguration,
    SingleNozzleLayoutConfiguration,
)
from opentrons.types import Point


@pytest.fixture
def available_sensors() -> AvailableSensorDefinition:
    """Provide a list of sensors."""
    return AvailableSensorDefinition(sensors=["pressure", "capacitive", "environment"])


@pytest.mark.parametrize(
    argnames=["request_model", "expected_nozzlemap", "nozzle_params"],
    argvalues=[
        [
            SingleNozzleLayoutConfiguration(primaryNozzle="A1"),
            NozzleMap.build(
                physical_nozzles=OrderedDict({"A1": Point(0, 0, 0)}),
                physical_rows=OrderedDict({"A": ["A1"]}),
                physical_columns=OrderedDict({"1": ["A1"]}),
                starting_nozzle="A1",
                back_left_nozzle="A1",
                front_right_nozzle="A1",
                valid_nozzle_maps=ValidNozzleMaps(maps={"A1": ["A1"]}),
            ),
            {"primary_nozzle": "A1"},
        ],
        [
            ColumnNozzleLayoutConfiguration(primaryNozzle="A1"),
            NozzleMap.build(
                physical_nozzles=NINETY_SIX_MAP,
                physical_rows=NINETY_SIX_ROWS,
                physical_columns=NINETY_SIX_COLS,
                starting_nozzle="A1",
                back_left_nozzle="A1",
                front_right_nozzle="H1",
                valid_nozzle_maps=ValidNozzleMaps(
                    maps={"Column1": NINETY_SIX_COLS["1"]}
                ),
            ),
            {"primary_nozzle": "A1", "front_right_nozzle": "H1"},
        ],
        [
            QuadrantNozzleLayoutConfiguration(
                primaryNozzle="A1", frontRightNozzle="E1", backLeftNozzle="A1"
            ),
            NozzleMap.build(
                physical_nozzles=NINETY_SIX_MAP,
                physical_rows=NINETY_SIX_ROWS,
                physical_columns=NINETY_SIX_COLS,
                starting_nozzle="A1",
                back_left_nozzle="A1",
                front_right_nozzle="E1",
                valid_nozzle_maps=ValidNozzleMaps(
                    maps={"A1_E1": ["A1", "B1", "C1", "D1", "E1"]}
                ),
            ),
            {"primary_nozzle": "A1", "front_right_nozzle": "E1"},
        ],
    ],
)
async def test_configure_nozzle_layout_implementation(
    decoy: Decoy,
    equipment: EquipmentHandler,
    tip_handler: TipHandler,
    available_sensors: AvailableSensorDefinition,
    request_model: Union[
        AllNozzleLayoutConfiguration,
        ColumnNozzleLayoutConfiguration,
        QuadrantNozzleLayoutConfiguration,
        SingleNozzleLayoutConfiguration,
    ],
    expected_nozzlemap: NozzleMap,
    nozzle_params: Dict[str, str],
) -> None:
    """A ConfigureForVolume command should have an execution implementation."""
    subject = ConfigureNozzleLayoutImplementation(
        equipment=equipment, tip_handler=tip_handler
    )

    config = LoadedStaticPipetteData(
        model="some-model",
        display_name="Hello",
        min_volume=0,
        max_volume=251,
        channels=8,
        home_position=123.1,
        nozzle_offset_z=331.0,
        flow_rates=FlowRates(
            default_aspirate={}, default_dispense={}, default_blow_out={}
        ),
        tip_configuration_lookup_table={},
        nominal_tip_overlap={},
        nozzle_map=get_default_nozzle_map(PipetteNameType.P300_MULTI),
        back_left_corner_offset=Point(10, 20, 30),
        front_right_corner_offset=Point(40, 50, 60),
        pipette_lld_settings={},
        plunger_positions={
            "top": 0.0,
            "bottom": 5.0,
            "blow_out": 19.0,
            "drop_tip": 20.0,
        },
        shaft_ul_per_mm=5.0,
        available_sensors=available_sensors,
        volume_mode=VolumeModes.lowVolumeDefault,
        available_volume_modes_min_vol={},
    )

    requested_nozzle_layout = ConfigureNozzleLayoutParams(
        pipetteId="pipette-id",
        configurationParams=request_model,
    )
    primary_nozzle = (
        None
        if isinstance(request_model, AllNozzleLayoutConfiguration)
        else request_model.primaryNozzle
    )
    front_right_nozzle = (
        request_model.frontRightNozzle
        if isinstance(request_model, QuadrantNozzleLayoutConfiguration)
        else None
    )
    back_left_nozzle = (
        request_model.backLeftNozzle
        if isinstance(request_model, QuadrantNozzleLayoutConfiguration)
        else None
    )

    decoy.when(
        await tip_handler.available_for_nozzle_layout(
            pipette_id="pipette-id",
            style=request_model.style,
            primary_nozzle=primary_nozzle,
            front_right_nozzle=front_right_nozzle,
            back_left_nozzle=back_left_nozzle,
        )
    ).then_return(nozzle_params)

    decoy.when(
        await equipment.configure_nozzle_layout(
            pipette_id="pipette-id",
            **nozzle_params,
        )
    ).then_return(
        LoadedConfigureNozzleLayoutData(
            pipette_id="pipette-id",
            serial_number="some number",
            nozzle_map=expected_nozzlemap,
            static_config=config,
        )
    )

    result = await subject.execute(requested_nozzle_layout)

    assert result == SuccessData(
        public=ConfigureNozzleLayoutResult(),
        state_update=StateUpdate(
            pipette_nozzle_map=PipetteNozzleMapUpdate(
                pipette_id="pipette-id",
                nozzle_map=expected_nozzlemap,
            ),
            pipette_config=update_types.PipetteConfigUpdate(
                pipette_id="pipette-id", serial_number="some number", config=config
            ),
        ),
    )



from typing import Optional, Protocol

from .types import LEDColor, LEDPattern, PressureState, PumpState, VacuumModuleInfo


class AbstractVacuumModuleDriver(Protocol):
    """Protocol for the Vacuum Module driver."""

    async def connect(self) -> None:
        """Connect to vacuum module."""
        ...

    async def disconnect(self) -> None:
        """Disconnect from vacuum module."""
        ...

    async def is_connected(self) -> bool:
        """Check connection to vacuum module."""
        ...

    async def get_device_info(self) -> VacuumModuleInfo:
        """Get Device Info."""
        ...

    async def set_serial_number(self, sn: str) -> None:
        """Set Serial Number."""
        ...

    async def enter_programming_mode(self) -> None:
        """Reboot into programming mode"""
        ...

    def reset_serial_buffers(self) -> None:
        """Reset the input and output serial buffers."""
        ...

    async def set_led(
        self,
        power: float,
        color: Optional[LEDColor] = None,
        external: Optional[bool] = None,
        pattern: Optional[LEDPattern] = None,
        duration: Optional[int] = None,  # Default firmware duration is 500ms
        reps: Optional[int] = None,  # Default firmware reps is 0
    ) -> None:
        """Set LED Status bar color and pattern."""
        ...

    async def set_vacuum_state(
        self,
        enable_vacuum: bool,
        guage_pressure_mbar: Optional[float] = None,
        duration: Optional[int] = None,
        rate: Optional[float] = None,
        vent_after: Optional[bool] = None,
    ) -> None:
        """Engage or release the vacuum until a desired internal pressure is reached."""
        ...

    async def get_vacuum_state(self) -> PressureState:
        """Get the pressure state."""
        ...

    async def set_pump_state(
        self,
        start_pump: bool,
        target_rpm: Optional[int] = None,
        duty_cycle: Optional[int] = None,
    ) -> None:
        """Start or the stop the pump at a given rpm or duty cycle."""
        ...

    async def get_pump_state(self) -> PumpState:
        """Get the pump state."""
        ...

    async def set_vent_state(self, state: bool) -> None:
        """Opens/Closes the vent, which release the vacuum in the module chamber."""
        ...



"""Test the tip overlap selection logic in the API core."""

import pytest

from ... import versions_at_or_above, versions_below
from opentrons.protocol_api.core.engine.overlap_versions import overlap_for_api_version
from opentrons.protocols.api_support.types import APIVersion


@pytest.mark.parametrize(
    "api_version", versions_below(APIVersion(2, 19), flex_only=False)
)
def test_all_below_219_use_v0(api_version: APIVersion) -> None:
    """Versions below 2.19 should use v0."""
    assert overlap_for_api_version(api_version) == "v0"


@pytest.mark.parametrize("api_version", versions_at_or_above(APIVersion(2, 19)))
def test_above_219_below_220_use_v1(api_version: APIVersion) -> None:
    """Versions above 2.19 and below 2.20 should use v1."""
    if api_version in versions_below(APIVersion(2, 20), flex_only=False):
        assert overlap_for_api_version(api_version) == "v1"


@pytest.mark.parametrize("api_version", versions_at_or_above(APIVersion(2, 20)))
def test_above_220_use_v3(api_version: APIVersion) -> None:
    """Versions above 2.20 should use v3."""
    assert overlap_for_api_version(api_version) == "v3"


def test_future_api_version_uses_v3() -> None:
    """Future versions should use v3."""
    assert overlap_for_api_version(APIVersion(2, 99)) == "v3"



"""Tests for ordered_set."""

import pytest

from opentrons.ordered_set import OrderedSet


def test_addition_and_iteration() -> None:
    """Elements should be returned in the order they were added."""
    subject = OrderedSet[int]()
    expected = list(range(100))
    for element in expected:
        subject.add(element)
    assert list(subject) == expected


def test_equality() -> None:
    """Sets should be equal only with the same elements in the same order."""
    subject_123 = OrderedSet[int]()
    subject_123.add(1)
    subject_123.add(2)
    subject_123.add(3)

    subject_123_also = OrderedSet[int]()
    subject_123_also.add(1)
    subject_123_also.add(2)
    subject_123_also.add(3)

    subject_123_float = OrderedSet[float]()
    subject_123_float.add(1.0)
    subject_123_float.add(2.0)
    subject_123_float.add(3.0)

    subject_123_str = OrderedSet[str]()
    subject_123_str.add("1")
    subject_123_str.add("2")
    subject_123_str.add("3")

    subject_456 = OrderedSet[int]()
    subject_456.add(4)
    subject_456.add(5)
    subject_456.add(6)

    subject_321 = OrderedSet[int]()
    subject_321.add(3)
    subject_321.add(2)
    subject_321.add(1)

    not_even_an_ordered_set = "💩"

    assert subject_123 == subject_123_also
    assert subject_123 == subject_123_float
    assert subject_123 != subject_123_str
    assert subject_123 != subject_456
    assert subject_123 != subject_321
    assert subject_123 != not_even_an_ordered_set


def test_deduplication() -> None:
    """Added elements should be deduplicated."""
    subject = OrderedSet[int]()
    for i in [3, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9]:
        subject.add(i)
    assert list(subject) == [3, 1, 4, 5, 9, 2, 6, 8, 7]


def test_initialization_by_iterable() -> None:
    """Initializing by iterable should be equivalent to adding one at a time."""
    source_iterable = [3, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9]
    subject_1 = OrderedSet[int](source_iterable)
    subject_2 = OrderedSet[int]()
    for element in source_iterable:
        subject_2.add(element)

    assert subject_1 == subject_2


def test_remove() -> None:
    """It should remove the element from the set, or raise if it's not present."""
    subject = OrderedSet([1, 2, 3, 4, 5])

    subject.remove(2)
    assert list(subject) == [1, 3, 4, 5]

    with pytest.raises(KeyError):
        subject.remove(2)
    assert list(subject) == [1, 3, 4, 5]


def test_discard() -> None:
    """It should remove the element from the set, or noop if it's not present."""
    subject = OrderedSet([1, 2, 3, 4, 5])

    subject.discard(2)
    assert list(subject) == [1, 3, 4, 5]

    subject.discard(2)  # Should just noop
    assert list(subject) == [1, 3, 4, 5]


def test_length() -> None:
    """The length should be the number of unique elements currently in the set."""
    subject = OrderedSet[str]()
    assert len(subject) == 0

    subject.add("a")
    assert len(subject) == 1

    subject.add("a")  # Duplicate.
    assert len(subject) == 1

    subject.add("b")
    assert len(subject) == 2

    subject.remove("a")
    assert len(subject) == 1


def test_clear() -> None:
    """It should clear all elements from the set."""
    subject = OrderedSet([1, 2, 3, 4, 5])
    subject.clear()
    assert list(subject) == []


def test_head() -> None:
    """It should return the head of the set."""
    subject = OrderedSet([1, 2])
    assert subject.head() == 1

    subject.remove(1)
    assert subject.head() == 2

    subject.remove(2)
    with pytest.raises(IndexError, match="Set is empty"):
        subject.head()

    assert subject.head(default_value=42) == 42


def test_difference() -> None:
    """It should return the set difference, preserving order."""
    a = OrderedSet([3, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5, 8, 9, 7, 9])
    b = {1, 9}

    assert (a - OrderedSet(b)) == (a - b) == OrderedSet([3, 4, 5, 2, 6, 5, 3, 5, 8, 7])


def test_repr() -> None:
    """It should return a meaningful repr string."""
    subject = OrderedSet([1, 2, 3])
    result = repr(subject)
    assert "OrderedSet" in result
    assert "1" in result
    assert "2" in result
    assert "3" in result



from typing import List, Optional, Tuple

import pytest
from pytest import MonkeyPatch

from opentrons.system import nmcli


def test_parse_colonsep() -> None:
    assert nmcli._parse_colonsep_response(
        "dank mimos\\: slow zone:100:yes:wpa2\n"
        "some other network:20:no:wep\n"
        "blah:4:yes:none\n"
    ) == [
        ["dank mimos: slow zone", "100", "yes", "wpa2"],
        ["some other network", "20", "no", "wep"],
        ["blah", "4", "yes", "none"],
    ]


def test_sanitize_args() -> None:
    cmd = [
        "nmcli",
        "connection",
        "add",
        "wifi.ssid",
        "Opentrons",
        "wifi-sec.psk",
        "test-password",
        "wifi-sec.key-mgmt",
        "wpa2-psk",
    ]
    sanitized = nmcli.sanitize_args(cmd)
    # Check preconditions
    assert "test-password" in cmd
    # Check output
    assert "test-password" not in sanitized

    cmd2 = ["nmcli", "connection", "modify", "+wifi-sec.psk", "test-password"]
    sanitized = nmcli.sanitize_args(cmd2)
    assert "test-password" in cmd2
    assert "test-password" not in sanitized


def test_output_transformations() -> None:
    fields = ["name", "type", "autorun", "active", "iface", "state"]
    should_have = [
        ["static-eth0", "802-3-ethernet", "yes", "yes", "eth0", "activated"],
        ["wifi-wlan0", "802-11-wireless", "yes", "no", "wlan0", "--"],
    ]
    # This test input is taken from the result of
    # nmcli -t -f name,type,autoconnect,active,device,state connection show
    test_input = """static-eth0:802-3-ethernet:yes:yes:eth0:activated
wifi-wlan0:802-11-wireless:yes:no:wlan0:--
"""
    # No transforms: correctly parse fields
    split = nmcli._dict_from_terse_tabular(fields, test_input)
    assert len(split) == 2
    for outp in zip(split, should_have):
        # All fields, in order, should be in the output
        assert fields == list(outp[0].keys())
        assert outp[1] == list(outp[0].values())

    # Transforms for some but not all keys
    transforms = {"name": lambda s: s.upper(), "active": lambda s: s == "yes"}
    split = nmcli._dict_from_terse_tabular(fields, test_input, transforms)
    assert split[0]["name"] == should_have[0][0].upper()
    assert split[1]["name"] == should_have[1][0].upper()
    assert split[0]["active"] is True
    assert split[1]["active"] is False


async def test_available_ssids(monkeypatch: MonkeyPatch) -> None:
    mock_nmcli_output = """mock_wpa2:90:no:WPA2
mock_no_security:80:no:
mock_enterprise:70:no:WPA1 WPA2 802.1X
mock_connected:60:yes:WPA2
mock_bad_security:50:no:foobar
--:40:no:"""

    expected_cmds = iter(
        (
            ["device", "wifi", "rescan"],
            [
                "--terse",
                "--fields",
                "ssid,signal,active,security",
                "device",
                "wifi",
                "list",
            ],
        )
    )

    expected = [
        {
            "ssid": "mock_wpa2",
            "signal": 90,
            "active": False,
            "security": "WPA2",
            "securityType": "wpa-psk",
        },
        {
            "ssid": "mock_no_security",
            "signal": 80,
            "active": False,
            "security": "",
            "securityType": "none",
        },
        {
            "ssid": "mock_enterprise",
            "signal": 70,
            "active": False,
            "security": "WPA1 WPA2 802.1X",
            "securityType": "wpa-eap",
        },
        {
            "ssid": "mock_connected",
            "signal": 60,
            "active": True,
            "security": "WPA2",
            "securityType": "wpa-psk",
        },
        {
            "ssid": "mock_bad_security",
            "signal": 50,
            "active": False,
            "security": "foobar",
            "securityType": "unsupported",
        },
        # note entry for 'ssid': '--' is expected to be filterd out
    ]

    async def mock_call(
        cmd: List[str], suppress_err: Optional[bool] = False
    ) -> Tuple[str, str]:
        assert cmd == next(expected_cmds)
        return mock_nmcli_output, ""

    monkeypatch.setattr(nmcli, "_call", mock_call)
    result = await nmcli.available_ssids(True)
    assert result == expected


async def test_networking_status(monkeypatch: MonkeyPatch) -> None:
    async def mock_call(cmd: List[str]) -> Tuple[str, str]:
        # Command: `nmcli networking connectivity`
        if "connectivity" in cmd:
            res = "full"
        elif "wlan0" in cmd:
            res = """GENERAL.HWADDR:B8:27:EB:5F:A6:89
IP4.ADDRESS[1]:--
IP4.GATEWAY:--
GENERAL.TYPE:wifi
GENERAL.STATE:30 (disconnected)"""
        elif "eth0" in cmd:
            res = """GENERAL.HWADDR:B8:27:EB:39:C0:9A
IP4.ADDRESS[1]:169.254.229.173/16
GENERAL.TYPE:ethernet
GENERAL.STATE:100 (connected)"""
        else:
            res = "incorrect nmcli call"

        return res, ""

    monkeypatch.setattr(nmcli, "_call", mock_call)

    assert await nmcli.is_connected() == "full"
    assert await nmcli.iface_info(nmcli.NETWORK_IFACES.WIFI) == {
        # test "--" gets mapped to None
        "ipAddress": None,
        "macAddress": "B8:27:EB:5F:A6:89",
        # test "--" gets mapped to None
        "gatewayAddress": None,
        "state": "disconnected",
        "type": "wifi",
    }

    assert await nmcli.iface_info(nmcli.NETWORK_IFACES.ETH_LL) == {
        "ipAddress": "169.254.229.173/16",
        "macAddress": "B8:27:EB:39:C0:9A",
        # test missing output gets mapped to None
        "gatewayAddress": None,
        "state": "connected",
        "type": "ethernet",
    }

    async def dummy_error_mock_call(cmd: List[str]) -> Tuple[str, str]:
        if "connectivity" in cmd:
            return "full", ""
        else:
            return "", "this is a dummy error"

    monkeypatch.setattr(nmcli, "_call", dummy_error_mock_call)
    assert await nmcli.is_connected() == "full"
    with pytest.raises(ValueError, match="this is a dummy error"):
        await nmcli.iface_info(nmcli.NETWORK_IFACES.WIFI)



"""Concurrency primitives providers."""

import asyncio


class ConcurrencyProvider:
    """Concurrency primitives for engine tasks."""

    def __init__(self) -> None:
        """Build a concurrency provider."""
        self._locks: dict[str, asyncio.Lock] = {}
        self._queues: dict[str, "asyncio.Queue[asyncio.Task[None]]"] = {}

    def lock_for_group(self, group_id: str) -> asyncio.Lock:
        """Returns the lock for specified group id."""
        try:
            return self._locks[group_id]
        except KeyError:
            self._locks[group_id] = asyncio.Lock()
            return self._locks[group_id]

    def queue_for_group(self, group_id: str) -> "asyncio.Queue[asyncio.Task[None]]":
        """Returns the queue for specified group id."""
        try:
            return self._queues[group_id]
        except KeyError:
            self._queues[group_id] = asyncio.Queue()
            return self._queues[group_id]



"""Test that the command schema is in sync with its source models."""

from opentrons_shared_data.command import get_newest_schema_version, load_schema_string

from opentrons.protocol_engine.commands import generate_command_schema

SCHEMA_CHANGED_MESSAGE = """\
The Python models for Protocol Engine commands do not match the shared JSON schema.

If this change is accidental, undo the changes to our Python models.

Or, if this change is intentional, update the shared JSON schema by running this from the monorepo root:

    make -C api command-schema
    make format

...and include the updated JSON schema file in your pull request.
"""


def test_command_schema_has_not_changed() -> None:
    """Confirm that the command schema has not changed."""
    newest_version = get_newest_schema_version()
    existing_schema_string = load_schema_string(newest_version)
    generated_schema_string = generate_command_schema(newest_version)

    assert generated_schema_string == existing_schema_string, SCHEMA_CHANGED_MESSAGE



"""Tests for the APIs around waste chutes and trash bins."""

import contextlib
import re
from typing import ContextManager, Optional, Type

import pytest

from opentrons import protocol_api
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import UnsupportedAPIError


@pytest.mark.parametrize(
    ("simulated_protocol_context", "expected_trash_class"),
    [
        (("2.13", "OT-2"), protocol_api.Labware),
        (("2.14", "OT-2"), protocol_api.Labware),
        (("2.15", "OT-2"), protocol_api.Labware),
        pytest.param(
            ("2.15", "Flex"),
            protocol_api.Labware,
            marks=pytest.mark.ot3_only,  # Simulating a Flex protocol requires a Flex hardware API.
        ),
        pytest.param(
            ("2.16", "OT-2"),
            protocol_api.TrashBin,
        ),
        pytest.param(
            ("2.16", "Flex"),
            None,
            marks=pytest.mark.ot3_only,  # Simulating a Flex protocol requires a Flex hardware API.
        ),
    ],
    indirect=["simulated_protocol_context"],
)
def test_fixed_trash_presence(
    simulated_protocol_context: protocol_api.ProtocolContext,
    expected_trash_class: Optional[Type[object]],
) -> None:
    """Test the presence of the fixed trash.

    Certain combinations of API version and robot type have a fixed trash.
    For those that do, ProtocolContext.fixed_trash and InstrumentContext.trash_container
    should point to it. The type of the object depends on the API version.
    """
    instrument = simulated_protocol_context.load_instrument(
        "p300_single_gen2"
        if simulated_protocol_context._core.robot_type == "OT-2 Standard"
        else "flex_1channel_50",
        mount="left",
    )

    if expected_trash_class is None:
        with pytest.raises(
            UnsupportedAPIError,
            match=re.escape(
                "Error 4002 API_REMOVED (UnsupportedAPIError): Fixed Trash is not available after API version 2.16."
                " You are currently using API version 2.16. Fixed trash is no longer supported on Flex protocols."
            ),
        ):
            simulated_protocol_context.fixed_trash
        with pytest.raises(Exception, match="No trash container has been defined"):
            instrument.trash_container

    else:
        assert isinstance(simulated_protocol_context.fixed_trash, expected_trash_class)
        assert instrument.trash_container is simulated_protocol_context.fixed_trash


@pytest.mark.ot3_only  # Simulating a Flex protocol requires a Flex hardware API.
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.16", "Flex")], indirect=True
)
def test_trash_search(simulated_protocol_context: protocol_api.ProtocolContext) -> None:
    """Test the automatic trash search for protocols without a fixed trash."""
    instrument = simulated_protocol_context.load_instrument(
        "flex_1channel_50", mount="left"
    )

    # By default, there should be no trash.
    with pytest.raises(
        UnsupportedAPIError,
        match=re.escape(
            "Error 4002 API_REMOVED (UnsupportedAPIError): Fixed Trash is not available after API version 2.16."
            " You are currently using API version 2.16. Fixed trash is no longer supported on Flex protocols."
        ),
    ):
        simulated_protocol_context.fixed_trash
    with pytest.raises(Exception, match="No trash container has been defined"):
        instrument.trash_container

    loaded_first = simulated_protocol_context.load_trash_bin("A1")
    loaded_second = simulated_protocol_context.load_trash_bin("B1")

    # After loading some trashes, there should still be no protocol.fixed_trash...
    with pytest.raises(
        UnsupportedAPIError,
        match=re.escape(
            "Error 4002 API_REMOVED (UnsupportedAPIError): Fixed Trash is not available after API version 2.16."
            " You are currently using API version 2.16. Fixed trash is no longer supported on Flex protocols."
        ),
    ):
        simulated_protocol_context.fixed_trash
    # ...but instrument.trash_container should automatically update to point to
    # the first trash that we loaded.
    assert instrument.trash_container is loaded_first

    # You should be able to override instrument.trash_container explicitly.
    instrument.trash_container = loaded_second
    assert instrument.trash_container is loaded_second


@pytest.mark.parametrize(
    ("simulated_protocol_context", "expect_load_to_succeed"),
    [
        pytest.param(
            ("2.13", "OT-2"),
            False,
            # This xfail (the system does let you load a labware onto slot 12, and does not raise)
            # is surprising to me. It may be be a bug in old PAPI versions.
            marks=pytest.mark.xfail(strict=True, raises=pytest.fail.Exception),
        ),
        (("2.14", "OT-2"), False),
        (("2.15", "OT-2"), False),
        pytest.param(
            ("2.15", "Flex"),
            False,
            marks=pytest.mark.ot3_only,  # Simulating a Flex protocol requires a Flex hardware API.
        ),
        pytest.param(
            ("2.16", "OT-2"),
            False,
        ),
        pytest.param(
            ("2.16", "Flex"),
            True,
            marks=pytest.mark.ot3_only,  # Simulating a Flex protocol requires a Flex hardware API.
        ),
    ],
    indirect=["simulated_protocol_context"],
)
def test_fixed_trash_load_conflicts(
    simulated_protocol_context: protocol_api.ProtocolContext,
    expect_load_to_succeed: bool,
) -> None:
    """Test loading something onto the location historically used for the fixed trash.

    In configurations where there is a fixed trash, this should be disallowed.
    In configurations without a fixed trash, this should be allowed.
    """
    if expect_load_to_succeed:
        expected_error: ContextManager[object] = contextlib.nullcontext()
    else:
        # If we're expecting an error, it'll be a LocationIsOccupied for 2.15 and below, otherwise
        # it will fail with an IncompatibleAddressableAreaError, since slot 12 will not be in the deck config
        if simulated_protocol_context.api_version < APIVersion(2, 16):
            error_name = "LocationIsOccupiedError"
        else:
            error_name = "IncompatibleAddressableAreaError"

        expected_error = pytest.raises(
            Exception,
            # Exact message doesn't matter, as long as it's definitely a labware load or addressable area conflict.
            match=error_name,
        )

    with expected_error:
        simulated_protocol_context.load_labware(
            "opentrons_96_wellplate_200ul_pcr_full_skirt", 12
        )



"""Tests for FileFormatValidator."""

import json

import pytest

from opentrons_shared_data import load_shared_data

from opentrons.protocol_reader.file_format_validator import (
    FileFormatValidationError,
    FileFormatValidator,
)
from opentrons.protocol_reader.file_identifier import (
    IdentifiedJsonMain,
    IdentifiedLabwareDefinition,
    IdentifiedPythonMain,
)
from opentrons.protocol_reader.file_reader_writer import BufferedFile
from opentrons.protocols.api_support.types import APIVersion


@pytest.mark.parametrize(
    "schema_version, json_protocol_fixture_path",
    [
        (3, "protocol/fixtures/3/simple.json"),
        (4, "protocol/fixtures/4/simpleV4.json"),
        (5, "protocol/fixtures/5/simpleV5.json"),
        (6, "protocol/fixtures/6/simpleV6.json"),
        (7, "protocol/fixtures/7/simpleV7.json"),
    ],
)
async def test_valid_json_main_file(
    schema_version: int, json_protocol_fixture_path: str
) -> None:
    """It should not raise when given a valid JSON main file."""
    json_protocol_contents = json.loads(load_shared_data(json_protocol_fixture_path))
    input_file = IdentifiedJsonMain(
        unvalidated_json=json_protocol_contents,
        schema_version=schema_version,
        # These attributes should be ignored by the subject:
        original_file=BufferedFile(
            name="foo",
            contents=b"bar",
            path=None,
        ),
        robot_type="OT-2 Standard",
        metadata={},
    )
    subject = FileFormatValidator()
    await subject.validate([input_file])


async def test_valid_python_main_file() -> None:
    """Python main files should always pass validation.

    FileFormatValidator doesn't currently enforce anything beyond the parsing done in
    FileIdentifier.
    """
    input_file = IdentifiedPythonMain(
        original_file=BufferedFile(
            name="foo",
            contents=b"bar",
            path=None,
        ),
        # These attributes should be ignored by the subject:
        api_level=APIVersion(123, 456),
        robot_type="OT-2 Standard",
        metadata={},
    )
    subject = FileFormatValidator()
    await subject.validate([input_file])


@pytest.mark.parametrize(
    "identified_schema_version",
    range(3, 8),
)
async def test_invalid_json_main(identified_schema_version: int) -> None:
    """It should raise when non-conforming JSON is provided as a main file."""
    input_file = IdentifiedJsonMain(
        unvalidated_json={"this_does_not_conform_to_any_opentrons_schema": True},
        schema_version=identified_schema_version,
        # These attributes should be ignored by the subject:
        original_file=BufferedFile(
            name="foo",
            contents=b"bar",
            path=None,
        ),
        robot_type="OT-2 Standard",
        metadata={},
    )
    subject = FileFormatValidator()
    with pytest.raises(
        FileFormatValidationError, match="foo could not be read as a JSON protocol."
    ):
        await subject.validate([input_file])


async def test_invalid_labware_definition() -> None:
    """It should raise when non-conforming JSON is provided as a labware file."""
    input_file = IdentifiedLabwareDefinition(
        original_file=BufferedFile(
            name="foo",
            contents=b"bar",
            path=None,
        ),
        unvalidated_json={"this_does_not_conform_to_any_opentrons_schema": True},
    )
    subject = FileFormatValidator()
    with pytest.raises(
        FileFormatValidationError,
        match="foo could not be read as a labware definition.",
    ):
        await subject.validate([input_file])



"""Test home commands."""

from decoy import Decoy

from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.home import (
    HomeImplementation,
    HomeParams,
    HomeResult,
)
from opentrons.protocol_engine.execution import MovementHandler
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.types import MotorAxis
from opentrons.types import MountType


async def test_home_implementation(decoy: Decoy, movement: MovementHandler) -> None:
    """A Home command should have an execution implementation."""
    subject = HomeImplementation(movement=movement)

    data = HomeParams(axes=[MotorAxis.X, MotorAxis.Y])

    result = await subject.execute(data)

    assert result == SuccessData(
        public=HomeResult(),
        state_update=update_types.StateUpdate(pipette_location=update_types.CLEAR),
    )
    decoy.verify(await movement.home(axes=[MotorAxis.X, MotorAxis.Y]))


async def test_home_all_implementation(decoy: Decoy, movement: MovementHandler) -> None:
    """It should pass axes=None along to the movement handler."""
    subject = HomeImplementation(movement=movement)

    data = HomeParams()

    result = await subject.execute(data)

    assert result == SuccessData(
        public=HomeResult(),
        state_update=update_types.StateUpdate(pipette_location=update_types.CLEAR),
    )
    decoy.verify(await movement.home(axes=None))


async def test_home_with_invalid_position(
    decoy: Decoy, movement: MovementHandler
) -> None:
    """Test handling when the homing is conditional (only move if position invalid)."""
    subject = HomeImplementation(movement=movement)

    decoy.when(
        await movement.check_for_valid_position(mount=MountType.LEFT)
    ).then_return(False)

    data = HomeParams(
        axes=[MotorAxis.X, MotorAxis.Y], skipIfMountPositionOk=MountType.LEFT
    )

    result = await subject.execute(data)
    assert result == SuccessData(
        public=HomeResult(),
        state_update=update_types.StateUpdate(pipette_location=update_types.CLEAR),
    )

    decoy.verify(await movement.home(axes=[MotorAxis.X, MotorAxis.Y]), times=1)
    decoy.reset()

    decoy.when(
        await movement.check_for_valid_position(mount=MountType.LEFT)
    ).then_return(True)
    result = await subject.execute(data)
    assert result == SuccessData(
        public=HomeResult(),
        state_update=update_types.StateUpdate(pipette_location=update_types.CLEAR),
    )

    decoy.verify(await movement.home(axes=[MotorAxis.X, MotorAxis.Y]), times=0)



"""Test fixtures for opentrons.protocol_runner tests.

These fixtures consist of two "matching" JSON and Python protocols,
saved to disk.
"""

import json
import textwrap
from pathlib import Path

import pytest

from opentrons_shared_data import load_shared_data
from opentrons_shared_data.labware import load_definition

from opentrons.protocol_engine import ModuleDefinition


@pytest.fixture(scope="session")
def tempdeck_v1_def() -> ModuleDefinition:
    """Get the definition of a V1 tempdeck."""
    definition = load_shared_data("module/definitions/3/temperatureModuleV1.json")
    return ModuleDefinition.model_validate_json(definition)


@pytest.fixture()
def json_protocol_file(tmp_path: Path) -> Path:
    """Get minimal JSON protocol input "file"."""
    tip_rack_def = load_definition("opentrons_96_tiprack_300ul", version=1)
    path = tmp_path / "protocol-name.json"

    path.write_text(
        json.dumps(
            {
                "$otSharedSchema": "#/protocol/schemas/6",
                "schemaVersion": 6,
                "metadata": {},
                "robot": {"model": "OT-2 Standard", "deckId": "ot2_standard"},
                "pipettes": {
                    "pipette-id": {"name": "p300_single"},
                },
                "labware": {
                    "labware-id": {
                        "displayName": "Opentrons 96 Tip Rack 300 µL",
                        "definitionId": "opentrons/opentrons_96_tiprack_300ul/1",
                    },
                },
                "labwareDefinitions": {
                    "opentrons/opentrons_96_tiprack_300ul/1": tip_rack_def,
                },
                "commands": [
                    {
                        "id": "command-id-1",
                        "commandType": "loadLabware",
                        "params": {
                            "labwareId": "labware-id",
                            "loadName": "opentrons_96_tiprack_300ul",
                            "namespace": "opentrons",
                            "version": 1,
                            "location": {"slotName": "1"},
                        },
                    },
                    {
                        "id": "command-id-2",
                        "commandType": "loadPipette",
                        "params": {
                            "pipetteId": "pipette-id",
                            "pipetteName": "p300_single",
                            "mount": "left",
                        },
                    },
                    {
                        "id": "command-id-3",
                        "commandType": "pickUpTip",
                        "params": {
                            "pipetteId": "pipette-id",
                            "labwareId": "labware-id",
                            "wellName": "A1",
                        },
                    },
                ],
            }
        )
    )

    return path


@pytest.fixture()
def python_protocol_file(tmp_path: Path) -> Path:
    """Get minimal Python protocol input "file"."""
    path = tmp_path / "protocol-name.py"
    path.write_text(
        textwrap.dedent(
            """
            # my protocol
            metadata = {
                "apiLevel": "2.14",
            }
            def run(ctx):
                pipette = ctx.load_instrument(
                    instrument_name="p300_single",
                    mount="left",
                )
                tip_rack = ctx.load_labware(
                    load_name="opentrons_96_tiprack_300ul",
                    location="1",
                )
                temp_module = ctx.load_module(
                    module_name="temperature module",
                    location="3"
                )
                pipette.pick_up_tip(
                    location=tip_rack.wells_by_name()["A1"],
                )
            """
        )
    )

    return path


@pytest.fixture()
def python_protocol_file_with_run_time_params(tmp_path: Path) -> Path:
    """Get minimal Python protocol input "file" with run time parameters."""
    path = tmp_path / "protocol-name.py"
    path.write_text(
        textwrap.dedent(
            """
            # my protocol
            metadata = {
                "apiLevel": "2.18",
            }
            def add_parameters(params):
                params.add_float(
                    display_name="Aspirate volume",
                    variable_name="aspirate_volume",
                    default=25.5,
                    minimum=10,
                    maximum=50,
                )
                params.add_str(
                    display_name="Mount",
                    variable_name="mount",
                    choices=[
                        {"display_name": "Left Mount", "value": "left"},
                        {"display_name": "Right Mount", "value": "right"},
                    ],
                    default="left",
                )
            def run(ctx):
                pipette = ctx.load_instrument(
                    instrument_name="p300_single",
                    mount=ctx.params.mount,
                )
                tip_rack = ctx.load_labware(
                    load_name="opentrons_96_tiprack_300ul",
                    location="1",
                )
                reservoir = ctx.load_labware(
                    load_name="nest_1_reservoir_195ml",
                    location=2,
                )
                pipette.pick_up_tip(
                    location=tip_rack.wells_by_name()["A1"],
                )
                pipette.aspirate(ctx.params.aspirate_volume, reservoir.wells()[0])
            """
        )
    )

    return path


@pytest.fixture()
def legacy_python_protocol_file(tmp_path: Path) -> Path:
    """Get an on-disk, minimal Python protocol fixture."""
    path = tmp_path / "protocol-name.py"
    path.write_text(
        textwrap.dedent(
            """
            # my protocol
            metadata = {
                "apiLevel": "2.11",
            }
            def run(ctx):
                pipette = ctx.load_instrument(
                    instrument_name="p300_single",
                    mount="left",
                )
                tip_rack = ctx.load_labware(
                    load_name="opentrons_96_tiprack_300ul",
                    location="1",
                )
                pipette.pick_up_tip(
                    location=tip_rack.wells_by_name()["A1"],
                )
            """
        )
    )

    return path


@pytest.fixture()
def legacy_json_protocol_file(tmp_path: Path) -> Path:
    """Get an on-disk, minimal JSON protocol fixture."""
    tip_rack_def = load_definition("opentrons_96_tiprack_300ul", version=1)
    path = tmp_path / "protocol-name.json"
    path.write_text(
        json.dumps(
            {
                "$otSharedSchema": "#/protocol/schemas/5",
                "schemaVersion": 5,
                "metadata": {},
                "robot": {"model": "OT-2 Standard"},
                "pipettes": {
                    "pipette-id": {"mount": "left", "name": "p300_single"},
                },
                "labware": {
                    "labware-id": {
                        "slot": "1",
                        "displayName": "Opentrons 96 Tip Rack 300 µL",
                        "definitionId": "opentrons/opentrons_96_tiprack_300ul/1",
                    },
                },
                "modules": {},
                "labwareDefinitions": {
                    "opentrons/opentrons_96_tiprack_300ul/1": tip_rack_def,
                },
                "commands": [
                    {
                        "command": "pickUpTip",
                        "params": {
                            "pipette": "pipette-id",
                            "labware": "labware-id",
                            "well": "A1",
                        },
                    },
                ],
            }
        )
    )

    return path



"""Test state getters for retrieving motion planning views of state."""

import inspect
from typing import List
from unittest.mock import sentinel

import pytest
from decoy import Decoy

from opentrons_shared_data.pipette.pipette_definition import (
    AvailableSensorDefinition,
)
from opentrons_shared_data.pipette.types import (
    LiquidClasses as VolumeModes,
)
from opentrons_shared_data.pipette.types import (
    PipetteNameType,
)
from opentrons_shared_data.robot.types import RobotType

from ..pipette_fixtures import get_default_nozzle_map
from opentrons import motion_planning
from opentrons.hardware_control.types import CriticalPoint
from opentrons.protocol_engine import errors
from opentrons.protocol_engine.state import _move_types
from opentrons.protocol_engine.state.addressable_areas import AddressableAreaView
from opentrons.protocol_engine.state.config import Config
from opentrons.protocol_engine.state.geometry import GeometryView
from opentrons.protocol_engine.state.labware import LabwareView
from opentrons.protocol_engine.state.module_substates import HeaterShakerModuleId
from opentrons.protocol_engine.state.modules import ModuleView
from opentrons.protocol_engine.state.motion import MotionView, PipetteLocationData
from opentrons.protocol_engine.state.pipettes import (
    BoundingNozzlesOffsets,
    PipetteBoundingBoxOffsets,
    PipetteView,
    StaticPipetteConfig,
)
from opentrons.protocol_engine.types import (
    AddressableOffsetVector,
    CurrentAddressableArea,
    CurrentWell,
    DeckSlotLocation,
    LoadedPipette,
    MotorAxis,
    WellLocation,
)
from opentrons.types import DeckSlotName, MountType, Point


@pytest.fixture
def mock_module_view(decoy: Decoy) -> ModuleView:
    """Get a mock in the shape of a ModuleView."""
    return decoy.mock(cls=ModuleView)


@pytest.fixture(autouse=True)
def patch_mock_get_waypoints(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    """Replace motion_planning.get_waypoints() with a mock."""
    mock_get_waypoints = decoy.mock(func=motion_planning.get_waypoints)
    monkeypatch.setattr(motion_planning, "get_waypoints", mock_get_waypoints)


@pytest.fixture(autouse=True)
def patch_mock__move_types(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    """Mock out _move_types.py functions."""
    for name, func in inspect.getmembers(_move_types, inspect.isfunction):
        monkeypatch.setattr(_move_types, name, decoy.mock(func=func))


def _fake_static_pipette_config(channels: int) -> StaticPipetteConfig:
    names_from_channels = {
        1: PipetteNameType.P1000_SINGLE_FLEX,
        8: PipetteNameType.P1000_MULTI_FLEX,
        96: PipetteNameType.P1000_96,
    }

    return StaticPipetteConfig(
        min_volume=1,
        max_volume=9001,
        channels=channels,
        model="blah",
        display_name="bleh",
        serial_number="",
        tip_configuration_lookup_table={},
        nominal_tip_overlap={},
        home_position=0,
        nozzle_offset_z=0,
        bounding_nozzle_offsets=BoundingNozzlesOffsets(
            back_left_offset=Point(x=10, y=20, z=30),
            front_right_offset=Point(x=40, y=50, z=60),
        ),
        default_nozzle_map=get_default_nozzle_map(names_from_channels[channels]),
        pipette_bounding_box_offsets=PipetteBoundingBoxOffsets(
            back_left_corner=Point(x=10, y=20, z=30),
            front_right_corner=Point(x=40, y=50, z=60),
            front_left_corner=Point(x=10, y=50, z=60),
            back_right_corner=Point(x=40, y=20, z=60),
        ),
        lld_settings={},
        plunger_positions={
            "top": 0.0,
            "bottom": 5.0,
            "blow_out": 19.0,
            "drop_tip": 20.0,
        },
        shaft_ul_per_mm=5.0,
        available_sensors=AvailableSensorDefinition(sensors=[]),
        volume_mode=VolumeModes.default,
        available_volume_modes_min_vol={},
    )


@pytest.fixture
def mock_engine_config(decoy: Decoy) -> Config:
    """Get a ProtocolEngine config value object."""
    return decoy.mock(cls=Config)


@pytest.fixture
def subject(
    mock_engine_config: Config,
    labware_view: LabwareView,
    pipette_view: PipetteView,
    addressable_area_view: AddressableAreaView,
    geometry_view: GeometryView,
    mock_module_view: ModuleView,
) -> MotionView:
    """Get a MotionView with its dependencies mocked out."""
    return MotionView(
        config=mock_engine_config,
        labware_view=labware_view,
        pipette_view=pipette_view,
        addressable_area_view=addressable_area_view,
        geometry_view=geometry_view,
        module_view=mock_module_view,
    )


def test_get_pipette_location_with_no_current_location(
    decoy: Decoy,
    pipette_view: PipetteView,
    subject: MotionView,
) -> None:
    """It should return mount and critical_point=None if no location."""
    decoy.when(pipette_view.get_current_location()).then_return(None)

    decoy.when(pipette_view.get("pipette-id")).then_return(
        LoadedPipette(
            id="pipette-id",
            mount=MountType.LEFT,
            pipetteName=PipetteNameType.P300_SINGLE,
        )
    )

    result = subject.get_pipette_location("pipette-id")

    assert result == PipetteLocationData(mount=MountType.LEFT, critical_point=None)


def test_get_pipette_location_with_current_location_with_y_center(
    decoy: Decoy,
    labware_view: LabwareView,
    pipette_view: PipetteView,
    subject: MotionView,
) -> None:
    """It should return cp=Y_CENTER if location labware requests."""
    decoy.when(pipette_view.get_current_location()).then_return(
        CurrentWell(pipette_id="pipette-id", labware_id="reservoir-id", well_name="A1")
    )

    decoy.when(pipette_view.get("pipette-id")).then_return(
        LoadedPipette(
            id="pipette-id",
            mount=MountType.RIGHT,
            pipetteName=PipetteNameType.P300_SINGLE,
        )
    )

    decoy.when(
        labware_view.get_should_center_column_on_target_well(
            "reservoir-id",
        )
    ).then_return(True)

    result = subject.get_pipette_location("pipette-id")

    assert result == PipetteLocationData(
        mount=MountType.RIGHT,
        critical_point=CriticalPoint.Y_CENTER,
    )


def test_get_pipette_location_with_current_location_with_xy_center(
    decoy: Decoy,
    labware_view: LabwareView,
    pipette_view: PipetteView,
    subject: MotionView,
) -> None:
    """It should return cp=XY_CENTER if location labware requests."""
    decoy.when(pipette_view.get_current_location()).then_return(
        CurrentWell(pipette_id="pipette-id", labware_id="reservoir-id", well_name="A1")
    )

    decoy.when(pipette_view.get("pipette-id")).then_return(
        LoadedPipette(
            id="pipette-id",
            mount=MountType.RIGHT,
            pipetteName=PipetteNameType.P300_SINGLE,
        )
    )

    decoy.when(
        labware_view.get_should_center_pipette_on_target_well(
            "reservoir-id",
        )
    ).then_return(True)

    result = subject.get_pipette_location("pipette-id")

    assert result == PipetteLocationData(
        mount=MountType.RIGHT,
        critical_point=CriticalPoint.XY_CENTER,
    )


def test_get_pipette_location_with_current_location_different_pipette(
    decoy: Decoy,
    labware_view: LabwareView,
    pipette_view: PipetteView,
    subject: MotionView,
) -> None:
    """It should return mount and cp=None if location used other pipette."""
    decoy.when(pipette_view.get_current_location()).then_return(
        CurrentWell(
            pipette_id="other-pipette-id",
            labware_id="reservoir-id",
            well_name="A1",
        )
    )

    decoy.when(pipette_view.get("pipette-id")).then_return(
        LoadedPipette(
            id="pipette-id",
            mount=MountType.LEFT,
            pipetteName=PipetteNameType.P300_SINGLE,
        )
    )

    decoy.when(
        labware_view.get_should_center_column_on_target_well(
            "reservoir-id",
        )
    ).then_return(False)

    decoy.when(
        labware_view.get_should_center_pipette_on_target_well(
            "reservoir-id",
        )
    ).then_return(False)

    result = subject.get_pipette_location("pipette-id")

    assert result == PipetteLocationData(
        mount=MountType.LEFT,
        critical_point=None,
    )


def test_get_pipette_location_override_current_location_xy_center(
    decoy: Decoy,
    labware_view: LabwareView,
    pipette_view: PipetteView,
    subject: MotionView,
) -> None:
    """It should calculate pipette location from a passed in deck location with xy override."""
    current_well = CurrentWell(
        pipette_id="pipette-id",
        labware_id="reservoir-id",
        well_name="A1",
    )

    decoy.when(pipette_view.get("pipette-id")).then_return(
        LoadedPipette(
            id="pipette-id",
            mount=MountType.RIGHT,
            pipetteName=PipetteNameType.P300_SINGLE,
        )
    )

    decoy.when(
        labware_view.get_should_center_pipette_on_target_well(
            "reservoir-id",
        )
    ).then_return(True)

    result = subject.get_pipette_location(
        pipette_id="pipette-id",
        current_location=current_well,
    )

    assert result == PipetteLocationData(
        mount=MountType.RIGHT,
        critical_point=CriticalPoint.XY_CENTER,
    )


def test_get_pipette_location_override_current_location_y_center(
    decoy: Decoy,
    labware_view: LabwareView,
    pipette_view: PipetteView,
    subject: MotionView,
) -> None:
    """It should calculate pipette location from a passed in deck location with xy override."""
    current_well = CurrentWell(
        pipette_id="pipette-id",
        labware_id="reservoir-id",
        well_name="A1",
    )

    decoy.when(pipette_view.get("pipette-id")).then_return(
        LoadedPipette(
            id="pipette-id",
            mount=MountType.RIGHT,
            pipetteName=PipetteNameType.P300_SINGLE,
        )
    )

    decoy.when(
        labware_view.get_should_center_column_on_target_well(
            "reservoir-id",
        )
    ).then_return(True)

    result = subject.get_pipette_location(
        pipette_id="pipette-id",
        current_location=current_well,
    )

    assert result == PipetteLocationData(
        mount=MountType.RIGHT,
        critical_point=CriticalPoint.Y_CENTER,
    )


@pytest.mark.parametrize("has_96_grid", [True, False])
@pytest.mark.parametrize("has_12_grid", [True, False])
@pytest.mark.parametrize("channels", [1, 8, 96])
def test_get_pipette_offset_for_reservoirs(
    decoy: Decoy,
    labware_view: LabwareView,
    pipette_view: PipetteView,
    geometry_view: GeometryView,
    mock_module_view: ModuleView,
    subject: MotionView,
    has_96_grid: bool,
    has_12_grid: bool,
    channels: int,
) -> None:
    """It should call get_waypoints() with the correct offset given the well's dimensions."""
    location = CurrentWell(pipette_id="123", labware_id="456", well_name="abc")
    decoy.when(
        geometry_view.get_min_travel_z("pipette-id", "labware-id", location, 123)
    ).then_return(42.0)

    decoy.when(pipette_view.get_current_location()).then_return(location)

    names_from_channels = {
        1: PipetteNameType.P1000_SINGLE_FLEX,
        8: PipetteNameType.P1000_MULTI_FLEX,
        96: PipetteNameType.P1000_96,
    }
    decoy.when(pipette_view.get_nozzle_configuration("pipette-id")).then_return(
        get_default_nozzle_map(names_from_channels[channels])
    )

    fake_x_dim, fake_y_dim, fake_z_dim = 7.0, 8.0, 9.0
    decoy.when(labware_view.get_well_size("labware-id", "well-name")).then_return(
        (fake_x_dim, fake_y_dim, fake_z_dim)
    )
    decoy.when(labware_view.get_has_96_subwells("labware-id")).then_return(has_96_grid)
    decoy.when(labware_view.get_has_12_subwells("labware-id")).then_return(has_12_grid)

    decoy.when(
        labware_view.get_should_center_column_on_target_well(
            "labware-id",
        )
    ).then_return(True)
    decoy.when(
        labware_view.get_should_center_pipette_on_target_well(
            "labware-id",
        )
    ).then_return(False)

    fake_well_position = Point(x=4, y=5, z=6)
    decoy.when(
        geometry_view.get_well_position(
            "labware-id", "well-name", WellLocation(), None, "pipette-id"
        )
    ).then_return(fake_well_position)

    decoy.when(
        _move_types.get_move_type_to_well(
            "pipette-id", "labware-id", "well-name", location, True
        )
    ).then_return(motion_planning.MoveType.GENERAL_ARC)
    decoy.when(
        geometry_view.get_min_travel_z("pipette-id", "labware-id", location, 123)
    ).then_return(42.0)

    decoy.when(geometry_view.get_ancestor_slot_name("labware-id")).then_return(
        DeckSlotName.SLOT_2
    )

    decoy.when(
        geometry_view.get_extra_waypoints(location, DeckSlotName.SLOT_2)
    ).then_return([(456, 789)])

    has_x_offset = has_y_offset = False
    if has_12_grid:
        if channels == 1 or channels == 8:
            has_x_offset = True
    if has_96_grid:
        if channels == 1:
            has_x_offset = has_y_offset = True
        if channels == 8:
            has_x_offset = True

    x_offset = -1 * fake_x_dim / 24 if has_x_offset else 0.0
    y_offset = fake_y_dim / 16 if has_y_offset else 0.0
    reservoir_offset = Point(x=x_offset, y=y_offset)
    expected_destination = fake_well_position + reservoir_offset

    # make sure get_waypoints is called with expected_destination
    decoy.when(
        motion_planning.get_waypoints(
            move_type=motion_planning.MoveType.GENERAL_ARC,
            origin=Point(x=1, y=2, z=3),
            origin_cp=CriticalPoint.MOUNT,
            max_travel_z=1337,
            min_travel_z=42,
            dest=expected_destination,
            dest_cp=CriticalPoint.Y_CENTER,
            xy_waypoints=[(456, 789)],
        )
    ).then_return(sentinel.waypoints)

    result = subject.get_movement_waypoints_to_well(
        pipette_id="pipette-id",
        labware_id="labware-id",
        well_name="well-name",
        well_location=WellLocation(),
        origin=Point(x=1, y=2, z=3),
        origin_cp=CriticalPoint.MOUNT,
        max_travel_z=1337,
        force_direct=True,
        minimum_z_height=123,
        offset_pipette_for_reservoir_subwells=True,
    )

    assert result is sentinel.waypoints


def test_get_movement_waypoints_to_well_for_y_center(
    decoy: Decoy,
    labware_view: LabwareView,
    pipette_view: PipetteView,
    geometry_view: GeometryView,
    mock_module_view: ModuleView,
    subject: MotionView,
) -> None:
    """It should call get_waypoints() with the correct args to move to a well."""
    location = CurrentWell(pipette_id="123", labware_id="456", well_name="abc")

    decoy.when(pipette_view.get_current_location()).then_return(location)

    decoy.when(
        labware_view.get_should_center_column_on_target_well(
            "labware-id",
        )
    ).then_return(True)
    decoy.when(
        labware_view.get_should_center_pipette_on_target_well(
            "labware-id",
        )
    ).then_return(False)

    decoy.when(
        geometry_view.get_well_position(
            "labware-id", "well-name", WellLocation(), None, "pipette-id"
        )
    ).then_return(Point(x=4, y=5, z=6))

    decoy.when(
        _move_types.get_move_type_to_well(
            "pipette-id", "labware-id", "well-name", location, True
        )
    ).then_return(motion_planning.MoveType.GENERAL_ARC)
    decoy.when(
        geometry_view.get_min_travel_z("pipette-id", "labware-id", location, 123)
    ).then_return(42.0)

    decoy.when(geometry_view.get_ancestor_slot_name("labware-id")).then_return(
        DeckSlotName.SLOT_2
    )
    decoy.when(pipette_view.get_config("pipette-id")).then_return(
        _fake_static_pipette_config(1)
    )

    decoy.when(
        geometry_view.get_extra_waypoints(location, DeckSlotName.SLOT_2)
    ).then_return([(456, 789)])

    waypoints = [
        motion_planning.Waypoint(
            position=Point(1, 2, 3), critical_point=CriticalPoint.Y_CENTER
        ),
        motion_planning.Waypoint(
            position=Point(4, 5, 6), critical_point=CriticalPoint.MOUNT
        ),
    ]

    decoy.when(
        motion_planning.get_waypoints(
            move_type=motion_planning.MoveType.GENERAL_ARC,
            origin=Point(x=1, y=2, z=3),
            origin_cp=CriticalPoint.MOUNT,
            max_travel_z=1337,
            min_travel_z=42,
            dest=Point(x=4, y=5, z=6),
            dest_cp=CriticalPoint.Y_CENTER,
            xy_waypoints=[(456, 789)],
        )
    ).then_return(waypoints)

    result = subject.get_movement_waypoints_to_well(
        pipette_id="pipette-id",
        labware_id="labware-id",
        well_name="well-name",
        well_location=WellLocation(),
        origin=Point(x=1, y=2, z=3),
        origin_cp=CriticalPoint.MOUNT,
        max_travel_z=1337,
        force_direct=True,
        minimum_z_height=123,
    )

    assert result == waypoints


def test_get_movement_waypoints_to_well_for_xy_center(
    decoy: Decoy,
    labware_view: LabwareView,
    pipette_view: PipetteView,
    geometry_view: GeometryView,
    mock_module_view: ModuleView,
    subject: MotionView,
) -> None:
    """It should call get_waypoints() with the correct args to move to a well."""
    location = CurrentWell(pipette_id="123", labware_id="456", well_name="abc")

    decoy.when(pipette_view.get_current_location()).then_return(location)

    decoy.when(
        labware_view.get_should_center_column_on_target_well(
            "labware-id",
        )
    ).then_return(False)
    decoy.when(
        labware_view.get_should_center_pipette_on_target_well(
            "labware-id",
        )
    ).then_return(True)

    decoy.when(
        geometry_view.get_well_position(
            "labware-id", "well-name", WellLocation(), None, "pipette-id"
        )
    ).then_return(Point(x=4, y=5, z=6))

    decoy.when(
        _move_types.get_move_type_to_well(
            "pipette-id", "labware-id", "well-name", location, True
        )
    ).then_return(motion_planning.MoveType.GENERAL_ARC)
    decoy.when(
        geometry_view.get_min_travel_z("pipette-id", "labware-id", location, 123)
    ).then_return(42.0)

    decoy.when(geometry_view.get_ancestor_slot_name("labware-id")).then_return(
        DeckSlotName.SLOT_2
    )

    decoy.when(
        geometry_view.get_extra_waypoints(location, DeckSlotName.SLOT_2)
    ).then_return([(456, 789)])

    waypoints = [
        motion_planning.Waypoint(
            position=Point(1, 2, 3), critical_point=CriticalPoint.XY_CENTER
        ),
        motion_planning.Waypoint(
            position=Point(4, 5, 6), critical_point=CriticalPoint.MOUNT
        ),
    ]

    decoy.when(
        motion_planning.get_waypoints(
            move_type=motion_planning.MoveType.GENERAL_ARC,
            origin=Point(x=1, y=2, z=3),
            origin_cp=CriticalPoint.MOUNT,
            max_travel_z=1337,
            min_travel_z=42,
            dest=Point(x=4, y=5, z=6),
            dest_cp=CriticalPoint.XY_CENTER,
            xy_waypoints=[(456, 789)],
        )
    ).then_return(waypoints)

    result = subject.get_movement_waypoints_to_well(
        pipette_id="pipette-id",
        labware_id="labware-id",
        well_name="well-name",
        well_location=WellLocation(),
        origin=Point(x=1, y=2, z=3),
        origin_cp=CriticalPoint.MOUNT,
        max_travel_z=1337,
        force_direct=True,
        minimum_z_height=123,
    )

    assert result == waypoints


def test_get_movement_waypoints_to_well_raises(
    decoy: Decoy,
    pipette_view: PipetteView,
    geometry_view: GeometryView,
    subject: MotionView,
) -> None:
    """It should raise FailedToPlanMoveError if motion_planning.get_waypoints raises."""
    decoy.when(
        geometry_view.get_well_position(
            labware_id="labware-id",
            well_name="A1",
            well_location=None,
            operation_volume=None,
            pipette_id="pipette-id",
        )
    ).then_return(Point(x=4, y=5, z=6))
    decoy.when(pipette_view.get_current_location()).then_return(None)
    decoy.when(
        geometry_view.get_min_travel_z("pipette-id", "labware-id", None, None)
    ).then_return(456)
    decoy.when(pipette_view.get_config("pipette-id")).then_return(
        _fake_static_pipette_config(1)
    )
    decoy.when(
        # TODO(mm, 2022-06-22): We should use decoy.matchers.Anything() for all
        # arguments. For some reason, Decoy does not match the call unless we
        # specify concrete values for all (?) arguments?
        motion_planning.get_waypoints(
            Point(x=1, y=2, z=3),
            Point(x=4, y=5, z=6),
            max_travel_z=123,
            min_travel_z=456,
        ),
        ignore_extra_args=True,
    ).then_raise(
        motion_planning.MotionPlanningError(
            origin=Point(1, 2, 3),
            dest=Point(1, 2, 3),
            clearance=123,
            min_travel_z=123,
            max_travel_z=123,
            message="oh the humanity",
        )
    )

    with pytest.raises(errors.FailedToPlanMoveError, match="oh the humanity"):
        subject.get_movement_waypoints_to_well(
            pipette_id="pipette-id",
            labware_id="labware-id",
            well_name="A1",
            well_location=None,
            origin=Point(1, 2, 3),
            origin_cp=None,
            max_travel_z=123,
        )


def test_get_movement_waypoints_to_addressable_area(
    decoy: Decoy,
    pipette_view: PipetteView,
    addressable_area_view: AddressableAreaView,
    geometry_view: GeometryView,
    subject: MotionView,
) -> None:
    """It should call get_waypoints() with the correct args to move to an addressable area."""
    location = CurrentAddressableArea(pipette_id="123", addressable_area_name="abc")

    decoy.when(pipette_view.get_current_location()).then_return(location)
    decoy.when(
        addressable_area_view.get_addressable_area_move_to_location("area-name")
    ).then_return(Point(x=3, y=3, z=3))
    decoy.when(geometry_view.get_all_obstacle_highest_z()).then_return(42)

    decoy.when(
        addressable_area_view.get_addressable_area_base_slot("area-name")
    ).then_return(DeckSlotName.SLOT_2)

    decoy.when(
        geometry_view.get_extra_waypoints(location, DeckSlotName.SLOT_2)
    ).then_return([])

    waypoints = [
        motion_planning.Waypoint(
            position=Point(1, 2, 3), critical_point=CriticalPoint.XY_CENTER
        ),
        motion_planning.Waypoint(
            position=Point(4, 5, 6), critical_point=CriticalPoint.MOUNT
        ),
    ]

    decoy.when(
        motion_planning.get_waypoints(
            move_type=motion_planning.MoveType.DIRECT,
            origin=Point(x=1, y=2, z=3),
            origin_cp=CriticalPoint.MOUNT,
            max_travel_z=1337,
            min_travel_z=123,
            dest=Point(x=4, y=5, z=6),
            dest_cp=CriticalPoint.XY_CENTER,
            xy_waypoints=[],
        )
    ).then_return(waypoints)

    result = subject.get_movement_waypoints_to_addressable_area(
        addressable_area_name="area-name",
        offset=AddressableOffsetVector(x=1, y=2, z=3),
        origin=Point(x=1, y=2, z=3),
        origin_cp=CriticalPoint.MOUNT,
        max_travel_z=1337,
        force_direct=True,
        minimum_z_height=123,
        ignore_tip_configuration=False,
    )

    assert result == waypoints


def test_move_to_moveable_trash_addressable_area(
    decoy: Decoy,
    pipette_view: PipetteView,
    addressable_area_view: AddressableAreaView,
    geometry_view: GeometryView,
    subject: MotionView,
) -> None:
    """Ensure that a move request to a moveableTrash addressable utilizes the Instrument Center critical point."""
    location = CurrentAddressableArea(
        pipette_id="123", addressable_area_name="moveableTrashA1"
    )

    decoy.when(pipette_view.get_current_location()).then_return(location)
    decoy.when(
        addressable_area_view.get_addressable_area_move_to_location("moveableTrashA1")
    ).then_return(Point(x=3, y=3, z=3))
    decoy.when(geometry_view.get_all_obstacle_highest_z()).then_return(42)

    decoy.when(
        addressable_area_view.get_addressable_area_base_slot("moveableTrashA1")
    ).then_return(DeckSlotName.SLOT_1)

    decoy.when(
        geometry_view.get_extra_waypoints(location, DeckSlotName.SLOT_1)
    ).then_return([])

    waypoints = [
        motion_planning.Waypoint(
            position=Point(1, 2, 3), critical_point=CriticalPoint.INSTRUMENT_XY_CENTER
        )
    ]

    decoy.when(
        motion_planning.get_waypoints(
            move_type=motion_planning.MoveType.DIRECT,
            origin=Point(x=1, y=2, z=3),
            origin_cp=CriticalPoint.MOUNT,
            max_travel_z=1337,
            min_travel_z=123,
            dest=Point(x=4, y=5, z=6),
            dest_cp=CriticalPoint.INSTRUMENT_XY_CENTER,
            xy_waypoints=[],
        )
    ).then_return(waypoints)

    result = subject.get_movement_waypoints_to_addressable_area(
        addressable_area_name="moveableTrashA1",
        offset=AddressableOffsetVector(x=1, y=2, z=3),
        origin=Point(x=1, y=2, z=3),
        origin_cp=CriticalPoint.MOUNT,
        max_travel_z=1337,
        force_direct=True,
        minimum_z_height=123,
        ignore_tip_configuration=True,
    )

    assert result == waypoints


def test_get_movement_waypoints_to_addressable_area_stay_at_max_travel_z(
    decoy: Decoy,
    pipette_view: PipetteView,
    addressable_area_view: AddressableAreaView,
    geometry_view: GeometryView,
    subject: MotionView,
) -> None:
    """It should call get_waypoints() with the correct args to move to an addressable area.

    This is the variant where we pass stay_at_max_travel_z=True to the subject.
    This should affect the dest argument of get_waypoints().
    """
    location = CurrentAddressableArea(pipette_id="123", addressable_area_name="abc")

    decoy.when(pipette_view.get_current_location()).then_return(location)
    decoy.when(
        addressable_area_view.get_addressable_area_move_to_location("area-name")
    ).then_return(Point(x=3, y=3, z=3))
    decoy.when(geometry_view.get_all_obstacle_highest_z()).then_return(42)

    decoy.when(
        addressable_area_view.get_addressable_area_base_slot("area-name")
    ).then_return(DeckSlotName.SLOT_2)

    decoy.when(
        geometry_view.get_extra_waypoints(location, DeckSlotName.SLOT_2)
    ).then_return([])

    waypoints = [
        motion_planning.Waypoint(
            position=Point(1, 2, 3), critical_point=CriticalPoint.XY_CENTER
        ),
        motion_planning.Waypoint(
            position=Point(4, 5, 6), critical_point=CriticalPoint.MOUNT
        ),
    ]

    decoy.when(
        motion_planning.get_waypoints(
            move_type=motion_planning.MoveType.DIRECT,
            origin=Point(x=1, y=2, z=3),
            origin_cp=CriticalPoint.MOUNT,
            max_travel_z=1337,
            min_travel_z=123,
            dest=Point(
                x=4,
                y=5,
                # The max_travel_z arg passed to the subject, plus the offset passed to the subject,
                # minus a 1 mm margin as a hack--see comments in the subject.
                z=1337 + 3 - 1,
            ),
            dest_cp=CriticalPoint.XY_CENTER,
            xy_waypoints=[],
        )
    ).then_return(waypoints)

    result = subject.get_movement_waypoints_to_addressable_area(
        addressable_area_name="area-name",
        offset=AddressableOffsetVector(x=1, y=2, z=3),
        origin=Point(x=1, y=2, z=3),
        origin_cp=CriticalPoint.MOUNT,
        max_travel_z=1337,
        force_direct=True,
        minimum_z_height=123,
        stay_at_max_travel_z=True,
        ignore_tip_configuration=False,
    )

    assert result == waypoints


@pytest.mark.parametrize(
    ("direct", "expected_move_type"),
    [
        (False, motion_planning.MoveType.GENERAL_ARC),
        (True, motion_planning.MoveType.DIRECT),
    ],
)
@pytest.mark.parametrize(
    ("additional_min_travel_z", "all_labware_highest_z", "expected_min_travel_z"),
    [
        (None, 100, 100),
        (200, 100, 200),
        (100, 200, 200),
        (None, -100, -100),
        (-200, -100, -100),
        (-100, -200, -100),
    ],
)
def test_get_movement_waypoints_to_coords(
    decoy: Decoy,
    geometry_view: GeometryView,
    subject: MotionView,
    direct: bool,
    expected_move_type: motion_planning.MoveType,
    additional_min_travel_z: float,
    all_labware_highest_z: float,
    expected_min_travel_z: float,
) -> None:
    """It should call get_waypoints() with the correct args to move to coordinates."""
    origin = Point(1, 2, 3)
    dest = Point(4, 5, 6)
    max_travel_z = 789

    decoy.when(geometry_view.get_all_obstacle_highest_z()).then_return(
        all_labware_highest_z
    )

    waypoints = [
        motion_planning.Waypoint(
            position=Point(1, 2, 3), critical_point=CriticalPoint.XY_CENTER
        ),
        motion_planning.Waypoint(
            position=Point(4, 5, 6), critical_point=CriticalPoint.MOUNT
        ),
    ]

    decoy.when(
        motion_planning.get_waypoints(
            origin=origin,
            origin_cp=None,
            dest=dest,
            dest_cp=None,
            min_travel_z=expected_min_travel_z,
            max_travel_z=max_travel_z,
            move_type=expected_move_type,
        )
    ).then_return(waypoints)

    result = subject.get_movement_waypoints_to_coords(
        origin=origin,
        dest=dest,
        max_travel_z=max_travel_z,
        direct=direct,
        additional_min_travel_z=additional_min_travel_z,
    )

    assert result == waypoints


def test_get_movement_waypoints_to_coords_raises(
    decoy: Decoy,
    geometry_view: GeometryView,
    subject: MotionView,
) -> None:
    """It should raise FailedToPlanMoveError if motion_planning.get_waypoints raises."""
    decoy.when(geometry_view.get_all_obstacle_highest_z()).then_return(123)
    decoy.when(
        # TODO(mm, 2022-06-22): We should use decoy.matchers.Anything() for all
        # arguments. For some reason, Decoy does not match the call unless we
        # specify concrete values for all (?) arguments?
        motion_planning.get_waypoints(
            Point(x=1, y=2, z=3),
            Point(x=1, y=2, z=3),
            max_travel_z=123,
            min_travel_z=123,
        ),
        ignore_extra_args=True,
    ).then_raise(
        motion_planning.MotionPlanningError(
            origin=Point(1, 2, 3),
            dest=Point(1, 2, 3),
            clearance=123,
            min_travel_z=123,
            max_travel_z=123,
            message="oh the humanity",
        )
    )

    with pytest.raises(errors.FailedToPlanMoveError, match="oh the humanity"):
        subject.get_movement_waypoints_to_coords(
            origin=Point(1, 2, 3),
            dest=Point(1, 2, 3),
            max_travel_z=123,
            direct=False,
            additional_min_travel_z=None,
        )


@pytest.mark.parametrize(
    ("labware_deck_slot", "expected_result"),
    [
        (DeckSlotName.SLOT_4, True),
        (DeckSlotName.SLOT_5, True),
        (DeckSlotName.SLOT_6, True),
        (DeckSlotName.SLOT_2, False),
        (DeckSlotName.SLOT_8, False),
        (DeckSlotName.SLOT_1, False),
    ],
)
def test_check_pipette_blocking_hs_latch(
    decoy: Decoy,
    geometry_view: GeometryView,
    pipette_view: PipetteView,
    mock_module_view: ModuleView,
    subject: MotionView,
    labware_deck_slot: DeckSlotName,
    expected_result: bool,
) -> None:
    """It should return True if pipette is blocking opening the latch."""
    decoy.when(pipette_view.get_current_location()).then_return(
        CurrentWell(pipette_id="pipette-id", labware_id="labware-id", well_name="A1")
    )

    decoy.when(geometry_view.get_ancestor_slot_name("labware-id")).then_return(
        labware_deck_slot
    )

    decoy.when(
        mock_module_view.get_location(HeaterShakerModuleId("heater-shaker-id"))
    ).then_return(DeckSlotLocation(slotName=DeckSlotName.SLOT_5))

    result = subject.check_pipette_blocking_hs_latch(
        HeaterShakerModuleId("heater-shaker-id")
    )

    assert result == expected_result


@pytest.mark.parametrize(
    ("labware_deck_slot", "expected_result"),
    [
        (DeckSlotName.SLOT_4, True),
        (DeckSlotName.SLOT_5, True),
        (DeckSlotName.SLOT_6, True),
        (DeckSlotName.SLOT_2, True),
        (DeckSlotName.SLOT_8, True),
        (DeckSlotName.SLOT_1, False),
    ],
)
def test_check_pipette_blocking_hs_shake(
    decoy: Decoy,
    geometry_view: GeometryView,
    pipette_view: PipetteView,
    mock_module_view: ModuleView,
    subject: MotionView,
    labware_deck_slot: DeckSlotName,
    expected_result: bool,
) -> None:
    """It should return True if pipette is blocking the h/s from shaking."""
    decoy.when(pipette_view.get_current_location()).then_return(
        CurrentWell(pipette_id="pipette-id", labware_id="labware-id", well_name="A1")
    )

    decoy.when(geometry_view.get_ancestor_slot_name("labware-id")).then_return(
        labware_deck_slot
    )

    decoy.when(
        mock_module_view.get_location(HeaterShakerModuleId("heater-shaker-id"))
    ).then_return(DeckSlotLocation(slotName=DeckSlotName.SLOT_5))

    result = subject.check_pipette_blocking_hs_shaker(
        HeaterShakerModuleId("heater-shaker-id")
    )

    assert result == expected_result


def test_get_touch_tip_waypoints(
    decoy: Decoy,
    labware_view: LabwareView,
    mock_module_view: ModuleView,
    geometry_view: GeometryView,
    pipette_view: PipetteView,
    subject: MotionView,
) -> None:
    """It should be able to get the position of a well top in a labware."""
    center_point = Point(1, 2, 3)

    decoy.when(
        labware_view.get_should_center_pipette_on_target_well("labware-id")
    ).then_return(True)
    decoy.when(
        labware_view.get_should_center_column_on_target_well("labware-id")
    ).then_return(False)

    decoy.when(pipette_view.get_mount("pipette-id")).then_return(MountType.LEFT)

    decoy.when(geometry_view.get_ancestor_slot_name("labware-id")).then_return(
        DeckSlotName.SLOT_4
    )

    decoy.when(
        mock_module_view.is_edge_move_unsafe(MountType.LEFT, DeckSlotName.SLOT_4)
    ).then_return(True)

    decoy.when(
        labware_view.get_edge_path_type(
            "labware-id", "B2", MountType.LEFT, DeckSlotName.SLOT_4, True
        )
    ).then_return(_move_types.EdgePathType.RIGHT)

    decoy.when(
        labware_view.get_well_radial_offsets("labware-id", "B2", 0.123)
    ).then_return((1.2, 3.4))

    decoy.when(
        _move_types.get_edge_point_list(
            center=center_point,
            x_radius=1.2,
            y_radius=3.4,
            edge_path_type=_move_types.EdgePathType.RIGHT,
            mm_from_edge=0.456,
        )
    ).then_return([Point(x=11, y=22, z=33), Point(x=44, y=55, z=66)])

    result = subject.get_touch_tip_waypoints(
        pipette_id="pipette-id",
        labware_id="labware-id",
        well_name="B2",
        center_point=center_point,
        radius=0.123,
        mm_from_edge=0.456,
    )

    assert result == [
        motion_planning.Waypoint(
            position=Point(x=11, y=22, z=33), critical_point=CriticalPoint.XY_CENTER
        ),
        motion_planning.Waypoint(
            position=Point(x=44, y=55, z=66), critical_point=CriticalPoint.XY_CENTER
        ),
    ]


@pytest.mark.parametrize(
    argnames=["robot_type", "expected_axes"],
    argvalues=[
        ["OT-2 Standard", [MotorAxis.LEFT_Z, MotorAxis.RIGHT_Z]],
        ["OT-3 Standard", [MotorAxis.LEFT_Z, MotorAxis.RIGHT_Z, MotorAxis.EXTENSION_Z]],
    ],
)
def test_get_mount_axes(
    decoy: Decoy,
    mock_engine_config: Config,
    subject: MotionView,
    robot_type: RobotType,
    expected_axes: List[MotorAxis],
) -> None:
    """It should return the correct mount axes for the robot type."""
    decoy.when(mock_engine_config.robot_type).then_return(robot_type)
    assert subject.get_robot_mount_axes() == expected_axes



"""Test pipette internal fluid tracking."""

import pytest

from opentrons.protocol_engine.state.fluid_stack import FluidStack
from opentrons.protocol_engine.types import AspiratedFluid, FluidKind


@pytest.mark.parametrize(
    "fluids,resulting_stack",
    [
        (
            [
                AspiratedFluid(FluidKind.LIQUID, 10),
                AspiratedFluid(FluidKind.LIQUID, 10),
            ],
            [AspiratedFluid(FluidKind.LIQUID, 20)],
        ),
        (
            [AspiratedFluid(FluidKind.AIR, 10), AspiratedFluid(FluidKind.LIQUID, 20)],
            [AspiratedFluid(FluidKind.AIR, 10), AspiratedFluid(FluidKind.LIQUID, 20)],
        ),
        (
            [
                AspiratedFluid(FluidKind.AIR, 10),
                AspiratedFluid(FluidKind.LIQUID, 20),
                AspiratedFluid(FluidKind.LIQUID, 10),
                AspiratedFluid(FluidKind.AIR, 20),
            ],
            [
                AspiratedFluid(FluidKind.AIR, 10),
                AspiratedFluid(FluidKind.LIQUID, 30),
                AspiratedFluid(FluidKind.AIR, 20),
            ],
        ),
    ],
)
def test_add_fluid(
    fluids: list[AspiratedFluid], resulting_stack: list[AspiratedFluid]
) -> None:
    """It should add fluids."""
    stack = FluidStack()
    for fluid in fluids:
        stack.add_fluid(fluid)
    assert stack._fluid_stack == resulting_stack


@pytest.mark.parametrize(
    "starting_fluids,remove_volume,resulting_stack",
    [
        ([], 1, []),
        ([], 0, []),
        (
            [AspiratedFluid(FluidKind.LIQUID, 10)],
            0,
            [AspiratedFluid(FluidKind.LIQUID, 10)],
        ),
        (
            [AspiratedFluid(FluidKind.LIQUID, 10)],
            5,
            [AspiratedFluid(FluidKind.LIQUID, 5)],
        ),
        ([AspiratedFluid(FluidKind.LIQUID, 10)], 11, []),
        (
            [AspiratedFluid(FluidKind.LIQUID, 10), AspiratedFluid(FluidKind.AIR, 10)],
            11,
            [AspiratedFluid(FluidKind.LIQUID, 9)],
        ),
        (
            [AspiratedFluid(FluidKind.LIQUID, 10), AspiratedFluid(FluidKind.AIR, 10)],
            20,
            [],
        ),
        (
            [
                AspiratedFluid(FluidKind.LIQUID, 10),
                AspiratedFluid(FluidKind.AIR, 10),
                AspiratedFluid(FluidKind.LIQUID, 10),
            ],
            28,
            [AspiratedFluid(FluidKind.LIQUID, 2)],
        ),
    ],
)
def test_remove_fluid(
    starting_fluids: list[AspiratedFluid],
    remove_volume: float,
    resulting_stack: list[AspiratedFluid],
) -> None:
    """It should remove fluids."""
    stack = FluidStack(_fluid_stack=[f for f in starting_fluids])
    stack.remove_fluid(remove_volume)
    assert stack._fluid_stack == resulting_stack


@pytest.mark.parametrize(
    "starting_fluids,filter,result",
    [
        ([], None, 0),
        ([], FluidKind.LIQUID, 0),
        ([], FluidKind.AIR, 0),
        ([AspiratedFluid(FluidKind.LIQUID, 10)], None, 10),
        ([AspiratedFluid(FluidKind.LIQUID, 10)], FluidKind.LIQUID, 10),
        ([AspiratedFluid(FluidKind.LIQUID, 10)], FluidKind.AIR, 0),
        (
            [AspiratedFluid(FluidKind.LIQUID, 10), AspiratedFluid(FluidKind.AIR, 10)],
            None,
            20,
        ),
        (
            [AspiratedFluid(FluidKind.LIQUID, 10), AspiratedFluid(FluidKind.AIR, 10)],
            FluidKind.LIQUID,
            10,
        ),
        (
            [AspiratedFluid(FluidKind.LIQUID, 10), AspiratedFluid(FluidKind.AIR, 10)],
            FluidKind.AIR,
            10,
        ),
    ],
)
def test_aspirated_volume(
    starting_fluids: list[AspiratedFluid], filter: FluidKind | None, result: float
) -> None:
    """It should represent aspirated volume with filtering."""
    stack = FluidStack(_fluid_stack=starting_fluids)
    assert stack.aspirated_volume(kind=filter) == result


@pytest.mark.parametrize(
    "starting_fluids,dispense_volume,result",
    [
        ([], 0, 0),
        ([], 1, 0),
        ([AspiratedFluid(FluidKind.AIR, 10)], 10, 0),
        ([AspiratedFluid(FluidKind.AIR, 10)], 0, 0),
        ([AspiratedFluid(FluidKind.LIQUID, 10)], 10, 10),
        ([AspiratedFluid(FluidKind.LIQUID, 10)], 0, 0),
        (
            [
                AspiratedFluid(FluidKind.AIR, 10),
                AspiratedFluid(FluidKind.LIQUID, 10),
            ],
            10,
            10,
        ),
        (
            [
                AspiratedFluid(FluidKind.AIR, 10),
                AspiratedFluid(FluidKind.LIQUID, 10),
            ],
            20,
            10,
        ),
        (
            [
                AspiratedFluid(FluidKind.AIR, 10),
                AspiratedFluid(FluidKind.LIQUID, 10),
            ],
            30,
            10,
        ),
        (
            [
                AspiratedFluid(FluidKind.AIR, 10),
                AspiratedFluid(FluidKind.LIQUID, 10),
            ],
            5,
            5,
        ),
        (
            [
                AspiratedFluid(FluidKind.LIQUID, 10),
                AspiratedFluid(FluidKind.AIR, 10),
            ],
            5,
            0,
        ),
        (
            [
                AspiratedFluid(FluidKind.LIQUID, 10),
                AspiratedFluid(FluidKind.AIR, 10),
            ],
            10,
            0,
        ),
        (
            [
                AspiratedFluid(FluidKind.LIQUID, 10),
                AspiratedFluid(FluidKind.AIR, 10),
            ],
            11,
            1,
        ),
        (
            [
                AspiratedFluid(FluidKind.LIQUID, 10),
                AspiratedFluid(FluidKind.AIR, 10),
            ],
            20,
            10,
        ),
        (
            [
                AspiratedFluid(FluidKind.LIQUID, 10),
                AspiratedFluid(FluidKind.AIR, 10),
            ],
            30,
            10,
        ),
    ],
)
def test_liquid_part_of_dispense_volume(
    starting_fluids: list[AspiratedFluid],
    dispense_volume: float,
    result: float,
) -> None:
    """It should predict resulting liquid from a dispense."""
    stack = FluidStack(_fluid_stack=starting_fluids)
    assert stack.liquid_part_of_dispense_volume(dispense_volume) == result



from opentrons.protocol_api.core.engine._default_labware_versions import (
    get_standard_labware_default_version,
)
from opentrons.protocol_engine.state.labware import LabwareLoadParams
from opentrons.protocols.api_support.constants import OPENTRONS_NAMESPACE
from opentrons.protocols.api_support.types import APIVersion


class AmbiguousLoadLabwareParamsError(RuntimeError):
    """Error raised when specific labware parameters cannot be found due to multiple matching labware definitions."""


def resolve(
    load_name: str,
    namespace: str | None,
    version: int | None,
    custom_load_labware_params: list[LabwareLoadParams],
    api_version: APIVersion,
) -> tuple[str, int]:
    """Resolve the load labware parameters that best matches any custom labware, or default to opentrons standards

    Args:
        load_name: Load name of the labware.
        namespace: Optionally provided labware definition namespace
        version: Optionally provided labware definition version
        custom_load_labware_params: List of load labware parameters associated with custom labware that
                                    match given parameters

    Returns:
        A tuple of the resolved namespace and version
    """

    def matches_params(custom_params: LabwareLoadParams) -> bool:
        matches_load_name = custom_params.load_name == load_name
        matches_namespace = namespace is None or custom_params.namespace == namespace
        matches_version = version is None or custom_params.version == version
        return matches_load_name and matches_namespace and matches_version

    if namespace is not None and version is not None:
        return namespace, version

    filtered_custom_params = [
        params for params in custom_load_labware_params if matches_params(params)
    ]

    if not filtered_custom_params:
        # No custom labware matches the input, but some standard labware might.
        # Use the Opentrons defaults for anything not explicitly provided.
        #
        # If the provided namespace was OPENTRONS_NAMESPACE, there would have been no
        # custom labware matching that namespace, so we will always take this path in
        # that case.
        resolved_namespace = namespace if namespace is not None else OPENTRONS_NAMESPACE
        resolved_version = (
            version
            if version is not None
            else get_standard_labware_default_version(
                load_name=load_name, api_version=api_version
            )
        )

    elif len(filtered_custom_params) > 1:
        # Multiple custom labware match the input.
        raise AmbiguousLoadLabwareParamsError(
            f"Multiple custom labware associated with load name {load_name}."
        )

    else:
        # Exactly one custom labware matches the input. Return it.
        resolved_namespace = filtered_custom_params[0].namespace
        resolved_version = filtered_custom_params[0].version

    return resolved_namespace, resolved_version



from opentrons import protocol_api, types

metadata = {
    "protocolName": "Testosaur",
    "author": "Opentrons <engineering@opentrons.com>",
    "description": 'A variant on "Dinosaur" for testing',
    "source": "Opentrons Repository",
    "apiLevel": "2.3",
}


def run(ctx: protocol_api.ProtocolContext) -> None:
    ctx.home()
    tr = ctx.load_labware("opentrons_96_tiprack_300ul", 1)
    right = ctx.load_instrument("p300_single_gen2", types.Mount.RIGHT, [tr])
    lw = ctx.load_labware("corning_96_wellplate_360ul_flat", 2)
    right.pick_up_tip()
    right.aspirate(10, lw.wells()[0].bottom())
    right.dispense(10, lw.wells()[1].bottom())
    right.drop_tip(tr.wells()[-1].top())



import pytest
from mock import AsyncMock

from opentrons.drivers.asyncio.communication.serial_connection import SerialConnection
from opentrons.drivers.command_builder import CommandBuilder
from opentrons.drivers.mag_deck.driver import (
    GCODE_ROUNDING_PRECISION,
    MAG_DECK_COMMAND_TERMINATOR,
    MagDeckDriver,
)


@pytest.fixture
def connection() -> AsyncMock:
    return AsyncMock(spec=SerialConnection)


@pytest.fixture
def driver(connection: AsyncMock) -> MagDeckDriver:
    connection.send_command.return_value = ""
    return MagDeckDriver(connection)


async def test_home(driver: MagDeckDriver, connection: AsyncMock) -> None:
    """It should send a home command"""
    await driver.home()

    expected = CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR).add_gcode(
        gcode="G28.2"
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)


async def test_probe_plate(driver: MagDeckDriver, connection: AsyncMock) -> None:
    """It should send a probe plate command"""
    await driver.probe_plate()

    expected = CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR).add_gcode(
        gcode="G38.2"
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)


async def test_get_plate_height(driver: MagDeckDriver, connection: AsyncMock) -> None:
    """It should send a get plate height command and parse response"""
    connection.send_command.return_value = "height:12.34"

    response = await driver.get_plate_height()

    expected = CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR).add_gcode(
        gcode="M836"
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)

    assert response == 12.34


async def test_get_mag_position(driver: MagDeckDriver, connection: AsyncMock) -> None:
    """It should send a get mag position command and parse response"""
    connection.send_command.return_value = "Z:12.34"

    response = await driver.get_mag_position()

    expected = CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR).add_gcode(
        gcode="M114.2"
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)

    assert response == 12.34


async def test_move(driver: MagDeckDriver, connection: AsyncMock) -> None:
    """It should send a move command"""
    await driver.move(321.2214)

    expected = (
        CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR)
        .add_gcode(gcode="G0")
        .add_float(prefix="Z", value=321.2214, precision=GCODE_ROUNDING_PRECISION)
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)


async def test_get_device_info(driver: MagDeckDriver, connection: AsyncMock) -> None:
    """It should send a get device info command and parse response"""
    connection.send_command.return_value = "serial:s model:m version:v"

    response = await driver.get_device_info()

    expected = CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR).add_gcode(
        gcode="M115"
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)

    assert response == {"serial": "s", "model": "m", "version": "v"}


async def test_enter_programming_mode(
    driver: MagDeckDriver, connection: AsyncMock
) -> None:
    """It should send an enter programming mode command"""
    await driver.enter_programming_mode()

    expected = CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR).add_gcode(
        gcode="dfu"
    )

    connection.send_command.assert_called_once_with(command=expected, retries=3)



import enum
from typing import List, cast

from opentrons.hardware_control import types


def test_create_aionotify_event() -> None:
    class FakeEnum(enum.Enum):
        CREATE = enum.auto()
        DELETE = enum.auto()
        MODIFY = enum.auto()

    enum_list: List[enum.Enum] = cast(
        List[enum.Enum], [FakeEnum.CREATE, FakeEnum.DELETE, FakeEnum.MODIFY]
    )

    new_event = types.AionotifyEvent.build("fake event", enum_list)

    assert hasattr(new_event.flags, "CREATE")
    assert hasattr(new_event.flags, "DELETE")
    assert hasattr(new_event.flags, "MODIFY")
    assert new_event.flags.CREATE
    assert new_event.flags.DELETE
    assert new_event.flags.MODIFY

    assert new_event.name == "fake event"



"""Test hardware stopping execution and side effects."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from decoy import Decoy

from opentrons.hardware_control import API as HardwareAPI
from opentrons.hardware_control.types import OT3Mount
from opentrons.protocol_engine.execution import (
    HardwareStopper,
    MovementHandler,
    TipHandler,
)
from opentrons.protocol_engine.state.state import StateStore
from opentrons.protocol_engine.types import (
    AddressableOffsetVector,
    MotorAxis,
    PostRunHardwareState,
    TipGeometry,
)
from opentrons.types import PipetteNotAttachedError as HwPipetteNotAttachedError

if TYPE_CHECKING:
    from opentrons.hardware_control.ot3api import OT3API


@pytest.fixture
def state_store(decoy: Decoy) -> StateStore:
    """Get a mocked out StateStore instance."""
    return decoy.mock(cls=StateStore)


@pytest.fixture
def hardware_api(decoy: Decoy) -> HardwareAPI:
    """Get a mocked out HardwareAPI instance."""
    return decoy.mock(cls=HardwareAPI)


@pytest.fixture
def movement(decoy: Decoy) -> MovementHandler:
    """Get a mocked out MovementHandler."""
    return decoy.mock(cls=MovementHandler)


@pytest.fixture
def mock_tip_handler(decoy: Decoy) -> TipHandler:
    """Get a mocked out TipHandler."""
    return decoy.mock(cls=TipHandler)


@pytest.fixture
def subject(
    hardware_api: HardwareAPI,
    state_store: StateStore,
    movement: MovementHandler,
    mock_tip_handler: TipHandler,
) -> HardwareStopper:
    """Get a HardwareStopper test subject with its dependencies mocked out."""
    return HardwareStopper(
        hardware_api=hardware_api,
        state_store=state_store,
        movement=movement,
        tip_handler=mock_tip_handler,
    )


async def test_hardware_halt(
    decoy: Decoy,
    hardware_api: HardwareAPI,
    subject: HardwareStopper,
) -> None:
    """It should halt the hardware API."""
    await subject.do_halt()

    decoy.verify(await hardware_api.halt(disengage_before_stopping=False))


@pytest.mark.parametrize(
    "post_run_hardware_state",
    [
        PostRunHardwareState.STAY_ENGAGED_IN_PLACE,
        PostRunHardwareState.DISENGAGE_IN_PLACE,
        PostRunHardwareState.HOME_AND_STAY_ENGAGED,
        PostRunHardwareState.HOME_THEN_DISENGAGE,
    ],
)
async def test_hardware_stopping_sequence(
    decoy: Decoy,
    state_store: StateStore,
    hardware_api: HardwareAPI,
    movement: MovementHandler,
    mock_tip_handler: TipHandler,
    subject: HardwareStopper,
    post_run_hardware_state: PostRunHardwareState,
) -> None:
    """It should stop the hardware, and home the robot. Flex no longer performs automatic drop tip.."""
    decoy.when(state_store.pipettes.get_all_attached_tips()).then_return(
        [
            ("pipette-id", TipGeometry(length=1.0, volume=2.0, diameter=3.0)),
        ]
    )
    decoy.when(hardware_api.attached_modules).then_return([])

    await subject.do_stop_and_recover(
        drop_tips_after_run=True,
        post_run_hardware_state=post_run_hardware_state,
    )

    decoy.verify(
        await hardware_api.stop(home_after=False),
        await movement.home(
            axes=[MotorAxis.X, MotorAxis.Y, MotorAxis.LEFT_Z, MotorAxis.RIGHT_Z]
        ),
        await hardware_api.stop(home_after=False),
    )


async def test_hardware_stopping_sequence_without_pipette_tips(
    decoy: Decoy,
    hardware_api: HardwareAPI,
    state_store: StateStore,
    subject: HardwareStopper,
    movement: MovementHandler,
) -> None:
    """Don't drop tip when there aren't any tips attached to pipettes."""
    decoy.when(state_store.pipettes.get_all_attached_tips()).then_return([])
    decoy.when(hardware_api.attached_modules).then_return([])

    await subject.do_stop_and_recover(
        drop_tips_after_run=True,
        post_run_hardware_state=PostRunHardwareState.HOME_AND_STAY_ENGAGED,
    )

    decoy.verify(
        await hardware_api.stop(home_after=False),
        await movement.home(
            [MotorAxis.X, MotorAxis.Y, MotorAxis.LEFT_Z, MotorAxis.RIGHT_Z]
        ),
    )


async def test_hardware_stopping_sequence_no_tip_drop(
    decoy: Decoy,
    state_store: StateStore,
    hardware_api: HardwareAPI,
    mock_tip_handler: TipHandler,
    subject: HardwareStopper,
) -> None:
    """Don't drop tip when told not to, even if tips are attached."""
    decoy.when(state_store.pipettes.get_all_attached_tips()).then_return(
        [
            ("pipette-id", TipGeometry(length=1.0, volume=2.0, diameter=3.0)),
        ]
    )
    decoy.when(hardware_api.attached_modules).then_return([])

    await subject.do_stop_and_recover(
        drop_tips_after_run=False,
        post_run_hardware_state=PostRunHardwareState.DISENGAGE_IN_PLACE,
    )

    decoy.verify(await hardware_api.stop(home_after=False), times=1)

    decoy.verify(
        mock_tip_handler.cache_tip(
            pipette_id="pipette-id",
            tip=TipGeometry(length=1.0, volume=2.0, diameter=3.0),
        ),
        times=0,
    )


async def test_hardware_stopping_sequence_no_pipette(
    decoy: Decoy,
    state_store: StateStore,
    hardware_api: HardwareAPI,
    mock_tip_handler: TipHandler,
    movement: MovementHandler,
    subject: HardwareStopper,
) -> None:
    """It should gracefully no-op if the HW API reports no attached pipette."""
    decoy.when(state_store.pipettes.get_all_attached_tips()).then_return(
        [
            ("pipette-id", TipGeometry(length=1.0, volume=2.0, diameter=3.0)),
        ]
    )

    decoy.when(
        mock_tip_handler.cache_tip(  # type: ignore[func-returns-value]
            pipette_id="pipette-id",
            tip=TipGeometry(length=1.0, volume=2.0, diameter=3.0),
        ),
    ).then_raise(HwPipetteNotAttachedError("oh no"))
    decoy.when(hardware_api.attached_modules).then_return([])

    await subject.do_stop_and_recover(
        drop_tips_after_run=True,
        post_run_hardware_state=PostRunHardwareState.HOME_AND_STAY_ENGAGED,
    )

    decoy.verify(
        await hardware_api.stop(home_after=False),
        await movement.home(
            [MotorAxis.X, MotorAxis.Y, MotorAxis.LEFT_Z, MotorAxis.RIGHT_Z]
        ),
        await hardware_api.stop(home_after=False),
        await movement.home(
            [MotorAxis.X, MotorAxis.Y, MotorAxis.LEFT_Z, MotorAxis.RIGHT_Z]
        ),
    )


@pytest.mark.ot3_only
async def test_hardware_stopping_sequence_with_gripper(
    decoy: Decoy,
    state_store: StateStore,
    ot3_hardware_api: OT3API,
    movement: MovementHandler,
    mock_tip_handler: TipHandler,
) -> None:
    """It should stop the hardware, and home the robot. Flex no longer performs automatic drop tip."""
    subject = HardwareStopper(
        hardware_api=ot3_hardware_api,
        state_store=state_store,
        movement=movement,
        tip_handler=mock_tip_handler,
    )
    decoy.when(state_store.pipettes.get_all_attached_tips()).then_return(
        [
            ("pipette-id", TipGeometry(length=1.0, volume=2.0, diameter=3.0)),
        ]
    )
    decoy.when(state_store.config.use_virtual_gripper).then_return(False)
    decoy.when(ot3_hardware_api.has_gripper()).then_return(True)
    decoy.when(subject._hardware_api.attached_modules).then_return([])

    await subject.do_stop_and_recover(
        drop_tips_after_run=True,
        post_run_hardware_state=PostRunHardwareState.HOME_AND_STAY_ENGAGED,
    )

    decoy.verify(
        await ot3_hardware_api.stop(home_after=False),
        await ot3_hardware_api.home_z(mount=OT3Mount.GRIPPER),
        await movement.home(
            axes=[MotorAxis.X, MotorAxis.Y, MotorAxis.LEFT_Z, MotorAxis.RIGHT_Z]
        ),
        await ot3_hardware_api.stop(home_after=False),
        await ot3_hardware_api.home_z(mount=OT3Mount.GRIPPER),
        await movement.home(
            axes=[MotorAxis.X, MotorAxis.Y, MotorAxis.LEFT_Z, MotorAxis.RIGHT_Z]
        ),
    )


@pytest.mark.ot3_only
async def test_hardware_stopping_sequence_with_fixed_trash(
    decoy: Decoy,
    state_store: StateStore,
    ot3_hardware_api: OT3API,
    movement: MovementHandler,
    mock_tip_handler: TipHandler,
) -> None:
    """It should stop the hardware, and home the robot. Flex no longer performs automatic drop tip."""
    subject = HardwareStopper(
        hardware_api=ot3_hardware_api,
        state_store=state_store,
        movement=movement,
        tip_handler=mock_tip_handler,
    )
    decoy.when(state_store.pipettes.get_all_attached_tips()).then_return(
        [
            ("pipette-id", TipGeometry(length=1.0, volume=2.0, diameter=3.0)),
        ]
    )
    decoy.when(state_store.labware.get_fixed_trash_id()).then_return("fixedTrash")
    decoy.when(state_store.config.use_virtual_gripper).then_return(False)
    decoy.when(ot3_hardware_api.has_gripper()).then_return(True)
    decoy.when(subject._hardware_api.attached_modules).then_return([])

    await subject.do_stop_and_recover(
        drop_tips_after_run=True,
        post_run_hardware_state=PostRunHardwareState.HOME_AND_STAY_ENGAGED,
    )

    decoy.verify(
        await ot3_hardware_api.stop(home_after=False),
        await ot3_hardware_api.home_z(mount=OT3Mount.GRIPPER),
        await movement.home(
            axes=[MotorAxis.X, MotorAxis.Y, MotorAxis.LEFT_Z, MotorAxis.RIGHT_Z]
        ),
        mock_tip_handler.cache_tip(
            pipette_id="pipette-id",
            tip=TipGeometry(length=1.0, volume=2.0, diameter=3.0),
        ),
        await movement.move_to_well(
            pipette_id="pipette-id",
            labware_id="fixedTrash",
            well_name="A1",
        ),
        await mock_tip_handler.drop_tip(
            pipette_id="pipette-id",
            home_after=False,
        ),
        await ot3_hardware_api.stop(home_after=False),
        await ot3_hardware_api.home_z(mount=OT3Mount.GRIPPER),
        await movement.home(
            axes=[MotorAxis.X, MotorAxis.Y, MotorAxis.LEFT_Z, MotorAxis.RIGHT_Z]
        ),
    )


async def test_hardware_stopping_sequence_with_OT2_addressable_area(
    decoy: Decoy,
    state_store: StateStore,
    hardware_api: HardwareAPI,
    movement: MovementHandler,
    mock_tip_handler: TipHandler,
) -> None:
    """It should stop the hardware, and home the robot. Flex no longer performs automatic drop tip."""
    subject = HardwareStopper(
        hardware_api=hardware_api,
        state_store=state_store,
        movement=movement,
        tip_handler=mock_tip_handler,
    )
    decoy.when(state_store.pipettes.get_all_attached_tips()).then_return(
        [
            ("pipette-id", TipGeometry(length=1.0, volume=2.0, diameter=3.0)),
        ]
    )
    decoy.when(state_store.config.robot_type).then_return("OT-2 Standard")
    decoy.when(state_store.config.use_virtual_gripper).then_return(False)
    decoy.when(hardware_api.attached_modules).then_return([])

    await subject.do_stop_and_recover(
        drop_tips_after_run=True,
        post_run_hardware_state=PostRunHardwareState.HOME_AND_STAY_ENGAGED,
    )

    decoy.verify(
        await hardware_api.stop(home_after=False),
        await movement.home(
            axes=[MotorAxis.X, MotorAxis.Y, MotorAxis.LEFT_Z, MotorAxis.RIGHT_Z]
        ),
        mock_tip_handler.cache_tip(
            pipette_id="pipette-id",
            tip=TipGeometry(length=1.0, volume=2.0, diameter=3.0),
        ),
        await movement.move_to_addressable_area(
            pipette_id="pipette-id",
            addressable_area_name="fixedTrash",
            offset=AddressableOffsetVector(x=0, y=0, z=0),
            force_direct=False,
            speed=None,
            minimum_z_height=None,
        ),
        await mock_tip_handler.drop_tip(
            pipette_id="pipette-id",
            home_after=False,
        ),
        await hardware_api.stop(home_after=False),
        await movement.home(
            axes=[MotorAxis.X, MotorAxis.Y, MotorAxis.LEFT_Z, MotorAxis.RIGHT_Z]
        ),
    )
